from ipaddress import ip_network
import re

import numpy as np
import pandas as pd
from ciscoconfparse import CiscoConfParse

from suzieq.engines.pandas.engineobj import SqPandasEngine
from suzieq.shared.confutils import (get_access_port_interfaces,
                                     get_trunk_port_interfaces)
from suzieq.shared.utils import build_query_str


class InterfacesObj(SqPandasEngine):
    '''Backend class to handle manipulating interfaces table with pandas'''

    @staticmethod
    def table_name():
        '''Table name'''
        return 'interfaces'

    def get(self, **kwargs):
        """Handling state outside of regular filters"""
        state = kwargs.pop('state', '')
        iftype = kwargs.pop('type', '')
        ifname = kwargs.get('ifname', '')
        vrf = kwargs.pop('vrf', '')
        master = kwargs.pop('master', [])
        columns = kwargs.pop('columns', [])
        user_query = kwargs.pop('query_str', '')
        vlan = kwargs.pop('vlan', '')
        portmode = kwargs.pop('portmode', '')

        addnl_fields = []
        if vrf:
            master.extend(vrf)

        fields = self.schema.get_display_fields(columns)
        self._add_active_to_fields(kwargs.get('view', self.iobj.view), fields,
                                   addnl_fields)

        drop_cols = []
        user_query_cols = self._get_user_query_cols(user_query)
        addnl_fields += [x for x in user_query_cols if x not in addnl_fields]

        if not ifname and iftype and iftype != ["all"]:
            df = super().get(type=iftype, master=master, columns=fields,
                             addnl_fields=addnl_fields, **kwargs)
        elif not ifname and iftype != ['all']:
            df = super().get(master=master, type=['!internal'], columns=fields,
                             addnl_fields=addnl_fields, **kwargs)
        else:
            df = super().get(master=master, columns=fields,
                             addnl_fields=addnl_fields, **kwargs)

        if df.empty:
            return df

        if portmode or any(x in fields
                           for x in ['vlan', 'vlanList', 'portmode']):
            for x in ['ipAddressList', 'ip6AddressList']:
                if x in columns or '*' in columns:
                    continue
                drop_cols.append(x)
            df = self._add_portmode(df, **kwargs)

        if vlan or "vlanList" in fields:
            df = self._add_vlanlist(df)

        if state or portmode:
            query_str = build_query_str([], self.schema, state=state,
                                        portmode=portmode)

            df = df.query(query_str)

        if vlan:
            # vlan needs to be looked at even in vlanList
            vlan = [int(x) for x in vlan]
            query_str = f' (vlan.isin({vlan}) or ' \
                f'@self._is_any_in_list(vlanList, {vlan}))'
            df = df.query(query_str)

        if user_query:
            df = self._handle_user_query_str(df, user_query)

        if not (iftype or ifname) and 'type' in df.columns:
            return df.query('type != "internal"') \
                     .reset_index(drop=True)[fields]
        else:
            return df.reset_index(drop=True)[fields]

    # pylint: disable=arguments-differ
    def aver(self, what="", **kwargs) -> pd.DataFrame:
        """Assert that interfaces are in good state"""

        ignore_missing_peer = kwargs.pop('ignore_missing_peer', False)

        if what == "mtu-value":
            result_df = self._assert_mtu_value(**kwargs)
        else:
            result_df = self._assert_interfaces(ignore_missing_peer, **kwargs)
        return result_df

    def summarize(self, **kwargs) -> pd.DataFrame:
        """Summarize interface information"""
        self._init_summarize(**kwargs)
        if self.summary_df.empty:
            return self.summary_df

        # Loopback interfaces on Linux have "unknown" as state
        self.summary_df["state"] = self.summary_df['state'] \
                                       .map({"unknown": "up",
                                             "up": "up", "down": "down",
                                             "notConnected": "notConnected"})

        self._summarize_on_add_field = [
            ('deviceCnt', 'hostname', 'nunique'),
            ('interfaceCnt', 'ifname', 'count'),
        ]

        self._summarize_on_add_with_query = [
            ('devicesWithL2Cnt', 'master == "bridge"', 'hostname', 'nunique'),
            ('devicesWithVxlanCnt', 'type == "vxlan"', 'hostname'),
            ('ifDownCnt', 'state != "up" and adminState == "up"', 'ifname'),
            ('ifAdminDownCnt', 'adminState != "up"', 'ifname'),
            ('ifWithMultipleIPCnt', 'ipAddressList.str.len() > 1', 'ifname'),
        ]

        self._summarize_on_add_list_or_count = [
            ('uniqueMTUCnt', 'mtu'),
            ('uniqueIfTypesCnt', 'type'),
            ('speedCnt', 'speed'),
        ]

        self._summarize_on_add_stat = [
            ('ifChangesStat', 'type != "bond"', 'numChanges'),
        ]

        self._summarize_on_perdevice_stat = [
            ('ifPerDeviceStat', '', 'ifname', 'count')
        ]

        self._gen_summarize_data()

        # The rest of the summary generation is too specific to interfaces
        original_summary_df = self.summary_df
        self.summary_df = original_summary_df.explode(
            'ipAddressList').dropna(how='any')

        if not self.summary_df.empty:
            self.nsgrp = self.summary_df.groupby(by=["namespace"])
            self._add_field_to_summary(
                'ipAddressList', 'nunique', 'uniqueIPv4AddrCnt')
        else:
            self._add_constant_to_summary('uniqueIPv4AddrCnt', 0)
        self.summary_row_order.append('uniqueIPv4AddrCnt')

        self.summary_df = original_summary_df \
            .explode('ip6AddressList') \
            .dropna(how='any') \
            .query('~ip6AddressList.str.startswith("fe80:")')

        if not self.summary_df.empty:
            self.nsgrp = self.summary_df.groupby(by=["namespace"])
            self._add_field_to_summary(
                'ip6AddressList', 'nunique', 'uniqueIPv6AddrCnt')
        else:
            self._add_constant_to_summary('uniqueIPv6AddrCnt', 0)
        self.summary_row_order.append('uniqueIPv6AddrCnt')

        self._post_summarize(check_empty_col='interfaceCnt')
        return self.ns_df.convert_dtypes()

    def _assert_mtu_value(self, **kwargs) -> pd.DataFrame:
        """Workhorse routine to match MTU value"""

        columns = ["namespace", "hostname", "ifname", "state", "mtu",
                   "timestamp"]

        matchval = kwargs.pop('matchval', [])
        result = kwargs.pop('result', '')

        matchval = [int(x) for x in matchval]

        result_df = self.get(columns=columns, **kwargs) \
                        .query('ifname != "lo"')

        if not result_df.empty:
            result_df['result'] = result_df.apply(
                lambda x, matchval: 'pass' if x['mtu'] in matchval else 'fail',
                axis=1, args=(matchval,))

        if result == "fail":
            result_df = result_df.query('result == "fail"')
        elif result == "pass":
            result_df = result_df.query('result == "pass"')

        return result_df

    # pylint: disable=too-many-statements
    def _assert_interfaces(self, ignore_missing_peer: bool, **kwargs) \
            -> pd.DataFrame:
        """Workhorse routine that validates MTU match for specified input"""
        columns = kwargs.pop('columns', [])
        result = kwargs.pop('result', 'all')
        state = kwargs.pop('state', '')
        iftype = kwargs.pop('type', [])

        def _check_field(x, fld1, fld2, reason):
            if x.skipIfCheck or x.indexPeer < 0:
                return []

            if x[fld1] == x[fld2]:
                return []
            return reason

        def _check_ipaddr(x, fld1, fld2, reason):
            # If we have no peer, don't check
            if x.skipIfCheck or x.indexPeer < 0:
                return []

            if len(x[fld1]) != len(x[fld2]):
                return reason

            if (len(x[fld1]) != 0):
                if (x[fld1][0].split('/')[1] == "32" or
                    (ip_network(x[fld1][0], strict=False) ==
                        ip_network(x[fld2][0], strict=False))):
                    return []
            else:
                return []

            return reason

        columns = ['*']

        if not state:
            state = 'up'

        if not iftype:
            iftype = ['ethernet', 'bond_slave', 'subinterface', 'vlan', 'bond']

        if_df = self.get(columns=columns, type=iftype, state=state, **kwargs)
        if if_df.empty:
            if result != 'pass':
                if_df['result'] = 'fail'
                if_df['assertReason'] = 'No data'

            return if_df

        if_df = if_df.drop(columns=['description', 'routeDistinguisher',
                                    'interfaceMac'], errors='ignore')

        # filter out loopback subinterfaces
        if 'loopback' not in iftype:
            lo_pattern = r'^(lo.*)|(.*oopback.*)$'
            if_df = if_df[if_df.apply(
                lambda x: re.match(lo_pattern, x.ifname) is None,
                axis=1)]

        if_df = self._drop_junos_pifnames(if_df).reset_index()

        lldpobj = self._get_table_sqobj('lldp')
        mlagobj = self._get_table_sqobj('mlag')

        # can't pass all kwargs, because lldp acceptable arguements are
        # different than interface
        namespace = kwargs.get('namespace', [])
        hostname = kwargs.get('hostname', [])
        lldp_df = lldpobj.get(namespace=namespace, hostname=hostname) \
                         .query('peerIfname != "-"')

        mlag_df = mlagobj.get(namespace=namespace, hostname=hostname)
        if not mlag_df.empty:
            mlag_peerlinks = set(mlag_df
                                 .groupby(by=['namespace', 'hostname',
                                              'peerLink'])
                                 .groups.keys())
        else:
            mlag_peerlinks = set()

        if 'vlanList' not in if_df.columns:
            if_df['vlanList'] = [[] for i in range(len(if_df))]

        if lldp_df.empty:
            if result != 'pass':
                if_df['assertReason'] = 'No LLDP peering info'
                if_df['result'] = 'fail'

            return if_df

        # Now create a single DF where you get the MTU for the lldp
        # combo of (namespace, hostname, ifname) and the MTU for
        # the combo of (namespace, peerHostname, peerIfname) and then
        # pare down the result to the rows where the two MTUs don't match
        idf = (
            pd.merge(
                if_df,
                lldp_df,
                left_on=["namespace", "hostname", "pifname"],
                right_on=['namespace', 'hostname', 'ifname'],
                how="outer",
            )
            .drop(columns=['ifname_y', 'timestamp_y'])
            .rename({'ifname_x': 'ifname', 'timestamp_x': 'timestamp',
                     'adminState_x': 'adminState',
                     'ipAddressList_x': 'ipAddressList',
                     'ip6AddressList_x': 'ip6AddressList',
                     'portmode_x': 'portmode'}, axis=1)
        )
        idf_nonsubif = idf.query('~type.isin(["subinterface", "vlan"])')
        idf_subif = idf.query('type.isin(["subinterface", "vlan"])')

        # Replace the bond_slave port interface with the bond interface

        idf_nonsubif = idf_nonsubif.merge(
            idf_nonsubif,
            left_on=["namespace", "peerHostname", "peerIfname"],
            right_on=['namespace', 'hostname', 'ifname'],
            how="outer", suffixes=["", "Peer"])

        idf_subif = idf_subif.merge(
            idf_subif,
            left_on=["namespace", "peerHostname", "peerIfname", 'vlan'],
            right_on=['namespace', 'hostname', 'pifname', 'vlan'],
            how="outer", suffixes=["", "Peer"])

        combined_df = pd.concat(
            [idf_subif, idf_nonsubif]).reset_index(drop=True)

        combined_df = combined_df \
            .drop(columns=["hostnamePeer", "pifnamePeer",
                           "mgmtIP", "description"]) \
            .dropna(subset=['hostname', 'ifname']) \
            .drop_duplicates(subset=['namespace', 'hostname', 'ifname'])

        if combined_df.empty:
            if result != 'pass':
                if_df['assertReason'] = 'No LLDP peering info'
                if_df['result'] = 'fail'

            return if_df

        combined_df = combined_df.fillna(
            {'mtuPeer': 0, 'speedPeer': 0, 'typePeer': '',
             'peerHostname': '', 'peerIfname': '', 'indexPeer': -1})
        for fld in ['ipAddressListPeer', 'ip6AddressListPeer', 'vlanListPeer']:
            combined_df[fld] = combined_df[fld] \
                .apply(lambda x: x if isinstance(x, np.ndarray) else [])

        combined_df['assertReason'] = combined_df.apply(
            lambda x: []
            if (x['adminState'] == 'down' or
                (x['adminState'] == "up" and x['state'] == "up"))
            else [x.reason or "Interface Down"], axis=1)

        known_hosts = set(combined_df.groupby(by=['namespace', 'hostname'])
                          .groups.keys())
        # Mark interfaces that can be skippedfrom checking because you cannot
        # find a peer
        combined_df['skipIfCheck'] = combined_df.apply(
            lambda x:
            (x.master == 'bridge') or (x.type in ['bond_slave', 'vlan']),
            axis=1)

        combined_df['indexPeer'] = combined_df.apply(
            lambda x, kh: x.indexPeer
            if (x.namespace, x.hostname) in kh else -2,
            args=(known_hosts,), axis=1)

        combined_df['assertReason'] += combined_df.apply(
            lambda x: ['No Peer Found']
            if x.indexPeer == -1 and not x.skipIfCheck else [],
            axis=1)

        combined_df['assertReason'] += combined_df.apply(
            lambda x: ['Unpolled Peer']
            if x.indexPeer == -2 and not x.skipIfCheck else [],
            axis=1)

        combined_df['assertReason'] += combined_df.apply(
            lambda x: _check_field(x, 'mtu', 'mtuPeer', ['MTU mismatch']),
            axis=1)

        combined_df['assertReason'] += combined_df.apply(
            lambda x: _check_field(
                x, 'speed', 'speedPeer', ['Speed mismatch']),
            axis=1)

        combined_df['assertReason'] += combined_df.apply(
            lambda x: []
            if (x.indexPeer < 0 or
                ((x['type'] == x['typePeer']) or
                 (x['type'] == 'vlan' and x['typePeer'] == 'subinterface') or
                    (x['type'].startswith('ether') and
                     x['typePeer'].startswith('ether'))))
            else ['type mismatch'],
            axis=1)

        combined_df['assertReason'] += combined_df.apply(
            lambda x: [] if (x.indexPeer < 0 or
                             (x['portmode'] == x['portmodePeer']))
            else ['portMode Mismatch'], axis=1)

        combined_df['assertReason'] += combined_df.apply(
            lambda x:
            _check_ipaddr(x, 'ipAddressList', 'ipAddressListPeer',
                          ['IP address mismatch']), axis=1)

        # We ignore MLAG peerlinks mainly because of NXOS erroneous output.
        # NXOS displays the VLANs associated with an interface via show vlan
        # which is then further pruned out by vPC. This pruned out list needs
        # to be extracted from the vPC output and used for the peerlink
        # instead of the output of show vlan for that interface. Since most
        # platforms perform their own MLAG consistency checks, we can skip
        # doing VLAN consistency check on the peerlink.
        # TODO: A better checker for MLAG peerlinks if needed at a later time.

        combined_df['assertReason'] += combined_df.apply(
            lambda x: [] if (((x.portmode in ['access', 'trunk']) and
                              (x.indexPeer < 0 or
                               (x['vlan'] == x['vlanPeer']))) or
                             (x.portmode in ['routed', 'unknown']))
            else ['pvid Mismatch'], axis=1)

        combined_df['assertReason'] += combined_df.apply(
            lambda x, mlag_peerlinks: []
            if ((x.indexPeer > 0 and
                ((x.namespace, x.hostname, x.master) not in mlag_peerlinks) and
                 (set(x['vlanList']) == set(x['vlanListPeer']))) or
                ((x.indexPeer < 0) or
                ((x.namespace, x.hostname, x.master) in mlag_peerlinks)))
            else ['VLAN set mismatch'], args=(mlag_peerlinks,), axis=1)

        if ignore_missing_peer:
            combined_df['result'] = combined_df.apply(
                lambda x: 'fail'
                if (len(x.assertReason) and
                    (x.assertReason[0] != 'No Peer Found'))
                else 'pass', axis=1)
        else:
            combined_df['result'] = combined_df.apply(
                lambda x: 'fail' if (len(x.assertReason)) else 'pass', axis=1)

        if result == "fail":
            combined_df = combined_df.query('result == "fail"').reset_index()
        elif result == "pass":
            combined_df = combined_df.query('result == "pass"').reset_index()

        combined_df['assertReason'] = combined_df['assertReason'].apply(
            lambda x: x if len(x) else '-'
        )

        return combined_df[['namespace', 'hostname', 'ifname', 'state',
                            'peerHostname', 'peerIfname', 'result',
                            'assertReason', 'timestamp']]

    def _drop_junos_pifnames(self, if_df: pd.DataFrame) -> pd.DataFrame:
        """This function drops parent interfaces of Junos subinterfaces ending
        with .0 and rename them as the parent interface

        Args:
            if_df (pd.DataFrame): interface dataframe

        Returns:
            pd.DataFrame: updated dataframe
        """
        # save the parent interface name in pifname column
        if_df['pifname'] = if_df.apply(
            lambda x: x['ifname'].split('.')[0]
            if x.type in ['subinterface', 'vlan']
            else x['ifname'], axis=1)

        # save the parent interface for .0 in a different column
        # this column will be used in the final merge
        if_df['pifname_0'] = if_df.apply(
            lambda x: x.ifname.split('.')[0] if x.ifname.endswith('.0')
            else x.ifname, axis=1
        )

        # Thanks for Junos, remove all the useless parent interfaces
        # if we have a .0 interface since thats the real deal

        # remove parent interfaces of .0 from dataframe
        parent_0_df = if_df.groupby(by=['namespace', 'hostname', 'pifname_0'])\
            .size().reset_index(name='counts')

        parent_0_df = parent_0_df[parent_0_df
                                  .apply(lambda x: x.counts > 1, axis=1)] \
            .rename(columns={'pifname_0': 'ifname'}) \
            .drop(columns=['counts'])

        res_df = pd.concat([if_df[['namespace', 'hostname', 'ifname']],
                            parent_0_df]).drop_duplicates(keep=False)

        if_df = if_df.merge(
            res_df,
            on=['namespace', 'hostname', 'ifname']
        )

        if_df['type'] = if_df.apply(lambda x: 'ethernet'
                                    if x['ifname'].endswith('.0')
                                    else x['type'], axis=1)

        # map subinterface into parent interface
        if_df['ifname'] = if_df.apply(
            lambda x: x['ifname'] if not x['ifname'].endswith('.0')
            else x['pifname'], axis=1)

        return if_df

    def _add_portmode(self, df: pd.DataFrame, **kwargs):
        """Add the switchport-mode i.e. acceess/trunk/routed'''

        :param df[pd.Dataframe]: The dataframe to add vlanList to
        :param kwargs[dict]: User specified dictionary of kwargs for host/ns
        :returns: original dataframe with portmode col added and filterd
        """

        if df.empty:
            return df

        namespace = kwargs.get('namespace', [])
        hostname = kwargs.get('hostname', [])

        conf_df = self._get_table_sqobj('devconfig') \
            .get(namespace=namespace, hostname=hostname)

        devdf = self._get_table_sqobj('device') \
            .get(namespace=namespace, hostname=hostname,
                 columns=['namespace', 'hostname', 'os', 'status', 'vendor'])

        pm_df = pd.DataFrame({'namespace': [], 'hostname': [],
                              'ifname': [], 'portmode': []})

        pm_list = []
        for row in conf_df.itertuples():
            # Check what type of device this is
            # TBD: SONIC support
            if not devdf.empty:
                nos = devdf[(devdf.namespace == row.namespace) &
                            (devdf.hostname == row.hostname)]['os'].tolist()
                if not nos:
                    continue

                nos = nos[0]
                if any(x in nos for x in ['junos', 'panos']):
                    syntax = 'junos'
                else:
                    # The way we pull out Cumulus Linux conf is also ios-like
                    syntax = 'ios'
            else:
                # Heuristics now
                if '\ninterfaces {\n' in row.config or \
                   'paloaltonetworks' in row.config:
                    syntax = 'junos'
                else:
                    syntax = 'ios'
            try:
                conf = CiscoConfParse(row.config.split('\n'), syntax=syntax)
            except Exception:  # pylint: disable=broad-except
                continue

            pm_dict = get_access_port_interfaces(conf, nos)
            pm_list.extend([{'namespace': row.namespace,
                             'hostname': row.hostname,
                             'ifname': k,
                             'portmode': 'access',
                             'vlan': v} for k, v in pm_dict.items()])
            pm_dict = get_trunk_port_interfaces(conf, nos)
            pm_list.extend([{'namespace': row.namespace,
                             'hostname': row.hostname,
                             'ifname': k,
                             'portmode': 'trunk',
                             'vlan': v} for k, v in pm_dict.items()])

        pm_df = pd.DataFrame(pm_list)

        df['portmode'] = np.where(df.ipAddressList.str.len() == 0,
                                  'unknown',
                                  'routed')
        df['portmode'] = np.where(df.ip6AddressList.str.len() != 0,
                                  'routed', df.portmode)

        if not pm_df.empty:
            df = df.merge(pm_df, how='left', on=[
                'namespace', 'hostname', 'ifname'],
                suffixes=['', '_y'])
            df['portmode'] = np.where(df.portmode_y.isnull(), df.portmode,
                                      df.portmode_y)

        df.loc[df.ifname == "bridge", 'portmode'] = ''
        if 'vlan_y' in df.columns:
            df['vlan'] = np.where(df.vlan_y.isnull(), df.vlan,
                                  df.vlan_y)
        df['vlan'] = df.vlan.astype(int)

        df['portmode'] = np.where(df.adminState != 'up', '',
                                  df.portmode)
        # handle EOS and other VXLAN ports which treat the interface
        # as a trunk port, as opposed to the access port mode of
        # the upto Cumulus 4.2.x Vxlan ports
        vxlan_ports = df.type == "vxlan"
        df.loc[vxlan_ports, 'portmode'] = df.loc[vxlan_ports] \
            .apply(lambda x: 'trunk' if x['portmode'] == 'routed'
                   else x['portmode'], axis=1)
        return df.drop(columns=['portmode_y', 'vlan_y'],
                       errors='ignore')

    def _add_vlanlist(self, df: pd.DataFrame, **kwargs) -> pd.DataFrame():
        """Add list of active, unpruned VLANs on trunked ports

        :param df[pd.Dataframe]: The dataframe to add vlanList to
        :param kwargs[dict]: User specified dictionary of kwargs for host/ns/if
        :returns: original dataframe with vlanList col added
        """

        if df.empty:
            return df

        vlan_df = self._get_table_sqobj('vlan') \
                      .get(namespace=kwargs.get('namespace', []))

        if vlan_df.empty:
            df['vlanList'] = [[] for _ in range(len(df))]
            return df

        # Transform the list of VLANs from VLAN-oriented to interface oriented
        vlan_if_df = vlan_df.explode('interfaces') \
                            .groupby(by=['namespace', 'hostname',
                                         'interfaces'])['vlan'].unique() \
                            .reset_index() \
                            .rename(columns={'interfaces': 'ifname',
                                             'vlan': 'vlanList'})

        vlan_if_df['vlanList'] = vlan_if_df.vlanList.apply(sorted)

        df = df.merge(vlan_if_df, how='left',
                      on=['namespace', 'hostname', 'ifname'])

        isnull = df.vlanList.isnull()
        df.loc[isnull, 'vlanList'] = pd.Series([[]] * isnull.sum()).values

        # Now remove the vlanList from all the access and routed ports
        # We leave them on the unknown ports because we may not have gotten
        # the config data to identify the port mode
        not_trunk = df.portmode.isin(["access", "routed"])
        df.loc[not_trunk, 'vlanList'] = pd.Series(
            [[]] * not_trunk.sum()).values

        return df



"""List of Exceptions specific to Suzieq, across all the modules."""


class NoLLdpError(Exception):
    """No LLDP error."""


class EmptyDataframeError(Exception):
    """Empty dataframe error."""


class PathLoopError(Exception):
    """Path loop error."""


class DBReadError(Exception):
    """Database read error."""


class DBNotFoundError(Exception):
    """Database not found error."""


class UserQueryError(Exception):
    """User query error."""


class UnknownDevtypeError(Exception):
    """Unknown dev type error."""


class SqPollerConfError(Exception):
    """Invalid poller configuration error."""


class InventorySourceError(Exception):
    """Unable to read or parse the inventory source."""


class PollingError(Exception):
    """Exception raised every time there is an error while polling"""


class SensitiveLoadError(Exception):
    """Sensitive informarmation isn't loaded correctly"""



import pandas as pd
from pandas.core.dtypes.dtypes import DatetimeTZDtype
import numpy as np

from suzieq.sqobjects.basicobj import SqObject
from suzieq.shared.utils import humanize_timestamp


class OspfObj(SqObject):
    '''The object providing access to the ospf table'''

    def __init__(self, **kwargs):
        super().__init__(table='ospf', **kwargs)
        self._valid_get_args = ['namespace', 'hostname', 'columns', 'area',
                                'vrf', 'ifname', 'state', 'query_str']
        self._valid_assert_args = self._valid_get_args + ['result']
        self._valid_arg_vals = {
            'state': ['full', 'other', 'passive', '!full', '!passive',
                      '!other', ''],
            'result': ['all', 'pass', 'fail'],
        }

    def humanize_fields(self, df: pd.DataFrame, _=None) -> pd.DataFrame:
        '''Humanize the timestamp and boot time fields'''
        if df.empty:
            return df

        if 'lastChangeTime' in df.columns:
            if not isinstance(df.lastChangeTime.dtype, DatetimeTZDtype):
                df['lastChangeTime'] = humanize_timestamp(
                    df.lastChangeTime.fillna(0),
                    self.cfg.get('analyzer', {}).get('timezone', None))

            if 'adjState' in df.columns:
                df['lastChangeTime'] = np.where(df.adjState == "passive",
                                                pd.Timestamp(0),
                                                df.lastChangeTime)

        return super().humanize_fields(df)

    def aver(self, **kwargs) -> pd.DataFrame:

        if kwargs.get('state', ''):
            raise ValueError('Cannot specify state with OSPF assert')

        return super().aver(**kwargs)



from suzieq.engines.pandas.engineobj import SqPandasEngine


class MacsObj(SqPandasEngine):
    '''Backend class to handle manipulating MAC table with pandas'''

    @staticmethod
    def table_name():
        '''Table name'''
        return 'macs'

    def get(self, **kwargs):
        if not self.iobj.table:
            raise NotImplementedError

        columns = kwargs.pop('columns', ['default'])
        moveCount = kwargs.pop('moveCount', None)
        view = kwargs.pop('view', self.iobj.view)
        remoteOnly = False
        localOnly = kwargs.pop('local', False)
        user_query = kwargs.pop('query_str', '')
        vtep = kwargs.get('remoteVtepIp', [])

        addnl_fields = []

        if vtep:
            if kwargs['remoteVtepIp'] == ['any']:
                del kwargs['remoteVtepIp']
                remoteOnly = True

        if 'moveCount' in columns or (columns == ['*']) or moveCount:
            if view == 'latest':
                view = 'hour'   # compute mac moves only for the last hour
            compute_moves = True
        else:
            compute_moves = False

        fields = self.schema.get_display_fields(columns)
        self._add_active_to_fields(view, fields, addnl_fields)

        user_query_cols = self._get_user_query_cols(user_query)
        addnl_fields += [x for x in user_query_cols if x not in addnl_fields]

        df = super().get(view=view, columns=fields, addnl_fields=addnl_fields,
                         **kwargs)

        if compute_moves and not df.empty:
            df = df.set_index('namespace hostname mackey macaddr'.split()) \
                   .sort_values(by=['timestamp'])

            # enhanced OIF, capturing both OIF and remoteVtep to capture moves
            # across local and remote
            df['eoif'] = df['oif'] + '-' + df['remoteVtepIp']
            # Check if eoif has changed in the next row. Since we're grouping
            # by the unique entries, OIF + nexthopIP should be only for the
            # provided index
            df['neoif'] = df.groupby(level=[0, 1, 2, 3])[
                'eoif'].shift(1).fillna(value=df['eoif'])
            df['moved'] = df.apply(
                lambda x: 1 if x.neoif != x.eoif else 0, axis=1)
            df['moveCount'] = df.groupby(level=[0, 1, 2, 3])[
                'moved'].cumsum()
            df = df.reset_index()

            if moveCount:
                try:
                    moveCount = int(moveCount)
                    df = df.query(
                        f'moveCount == {moveCount}').reset_index(drop=True)
                except ValueError:
                    df = df.query(
                        f'moveCount {moveCount}').reset_index(drop=True)

        df = self._handle_user_query_str(df, user_query)
        if remoteOnly:
            df = df.query("remoteVtepIp != ''").reset_index(drop=True)
        elif localOnly:
            df = df.query("remoteVtepIp == ''").reset_index(drop=True)

        if columns in [['default'], ['*']] or 'mackey' not in columns:
            if 'mackey' in fields:
                fields.remove('mackey')

        return df[fields]

    def summarize(self, **kwargs):
        """Summarize the MAC table info"""

        self._summarize_on_add_field = [
            ('deviceCnt', 'hostname', 'nunique'),
            ('totalMacsinNSCnt', 'hostname', 'count'),
            ('uniqueMacCnt', 'macaddr', 'nunique'),
        ]

        self._summarize_on_perdevice_stat = [
            ('uniqueVlanperHostStat', 'vlan != 0 and vlan != ""', 'vlan',
             'nunique'),
        ]

        return super().summarize(**kwargs)



#!/usr/bin/env python3
"""
This module contains the logic needed to start the poller
"""

import argparse
import asyncio
import sys
import traceback
from typing import Dict

import uvloop
from suzieq.poller.controller.controller import Controller
from suzieq.poller.worker.writers.output_worker import OutputWorker
from suzieq.shared.exceptions import InventorySourceError, PollingError, \
    SqPollerConfError
from suzieq.shared.utils import (poller_log_params, init_logger,
                                 load_sq_config, print_version)
from suzieq.poller.controller.utils.inventory_utils import read_inventory


async def start_controller(user_args: argparse.Namespace, config_data: Dict):
    """Start the poller, this function launches the controller which
    spawns the poller workers

    Args:
        user_args (argparse.Namespace): the command line arguments
        config_data (Dict): the content of the Suzieq configuration file
    """
    # Init logger of the poller
    logfile, loglevel, logsize, log_stdout = poller_log_params(
        config_data,
        is_controller=True
    )
    logger = init_logger('suzieq.poller.controller', logfile,
                         loglevel, logsize, log_stdout)

    try:
        if user_args.syntax_check:
            if not user_args.inventory:
                raise SqPollerConfError(
                    'With the --syntax-check option the user must specify '
                    'the inventory file via the -I option')
            # perform inventory validation and return
            read_inventory(user_args.inventory)
            print('Inventory syntax check passed')
            return
        controller = Controller(user_args, config_data)
        controller.init()
        await controller.run()
    except (SqPollerConfError, InventorySourceError, PollingError) as error:
        if not log_stdout:
            print(f"ERROR: {error}")
        logger.error(error)
        sys.exit(-1)


def controller_main():
    """The routine that kicks things off including arg parsing
    """
    parser = argparse.ArgumentParser()

    # Get supported output, 'gather' cannot be manually selected
    supported_outputs = OutputWorker.get_plugins()
    if supported_outputs.get('gather', None):
        del supported_outputs['gather']
    supported_outputs = list(supported_outputs)

    # Two inputs are possible:
    # 1. Suzieq inventory file
    # 2. Input directory
    source_arg = parser.add_mutually_exclusive_group()
    source_arg.add_argument(
        '-I',
        '--inventory',
        type=str,
        help='Input inventory file'
    )

    source_arg.add_argument(
        '-i',
        '--input-dir',
        type=str,
        help=('Directory where run-once=gather data is. Process the data in '
              'directory as they were retrieved by the hosts')
    )

    parser.add_argument(
        '-c',
        '--config',
        help='Controller configuration file',
        type=str
    )

    parser.add_argument(
        '--debug',
        action='store_true',
        help='Build the node list and exit without polling the nodes'
    )

    parser.add_argument(
        '-x',
        '--exclude-services',
        type=str,
        help='Exclude running this space separated list of services',
    )

    parser.add_argument(
        '--no-coalescer',
        default=False,
        action='store_true',
        help='Do not start the coalescer',
    )

    parser.add_argument(
        '-o',
        '--outputs',
        nargs='+',
        default=['parquet'],
        choices=supported_outputs,
        type=str,
        help='Output formats to write to: parquet. Use '
        'this option multiple times for more than one output',
    )

    parser.add_argument(
        "--output-dir",
        type=str,
        help=argparse.SUPPRESS,
    )

    parser.add_argument(
        '--run-once',
        type=str,
        choices=['gather', 'process', 'update'],
        help=('''The poller do not run forever, three modes are available:
        (1) gather: store the output as it has been collected,
        (2) process: performs some processing on the data.
        Both cases store the results in a plain output file,
        one for each service, and exit.
        (3) update: poll the nodes only once, write the result and stop''')
    )

    parser.add_argument(
        '-s',
        '--service-only',
        type=str,
        help='Only run this space separated list of services',
    )

    parser.add_argument(
        '--ssh-config-file',
        type=str,
        default=None,
        help='Path to ssh config file, that you want to use'
    )

    parser.add_argument(
        '-p',
        '--update-period',
        help='How frequently the inventory updates [DEFAULT=3600]',
        type=int
    )

    parser.add_argument(
        '-w',
        '--workers',
        type=int,
        help='The number of workers polling the nodes',
    )

    parser.add_argument(
        '-V',
        '--version',
        action='store_true',
        help='Print suzieq version'
    )

    parser.add_argument(
        '--syntax-check',
        action='store_true',
        help='Check inventory file syntax and return'
    )

    args = parser.parse_args()

    if args.version:
        print_version()
        sys.exit(0)

    uvloop.install()
    cfg = load_sq_config(config_file=args.config)
    if not cfg:
        print("Could not load config file, aborting")
        sys.exit(1)

    try:
        asyncio.run(start_controller(args, cfg))
    except (KeyboardInterrupt, RuntimeError):
        pass
    except Exception:  # pylint: disable=broad-except
        traceback.print_exc()


if __name__ == '__main__':
    controller_main()



#!/usr/bin/env python3

import argparse
import asyncio
import os
import sys
import traceback
from typing import Dict

import uvloop

from suzieq.poller.worker.worker import Worker
from suzieq.poller.worker.writers.output_worker import OutputWorker
from suzieq.shared.exceptions import InventorySourceError, SqPollerConfError
from suzieq.shared.utils import poller_log_params, init_logger, load_sq_config


async def start_worker(userargs: argparse.Namespace, cfg: Dict):
    """Start the poller worker

    Args:
        userargs (argparse.Namespace): the command line arguments
        cfg (Dict): the content of the Suzieq config file
    """
    # Init logger of the poller
    logfile, loglevel, logsize, log_stdout = poller_log_params(
        cfg,
        worker_id=userargs.worker_id
    )
    logger = init_logger('suzieq.poller.worker', logfile,
                         loglevel, logsize, log_stdout)

    worker = None
    try:
        worker = Worker(userargs, cfg)
        await worker.init_worker()
        await worker.run()
    except (SqPollerConfError, InventorySourceError) as error:
        logger.error(error)
        print(error)
        sys.exit(1)


def worker_main():
    """The routine that kicks things off including arg parsing
    """
    # Get supported output, 'gather' cannot be manually selected
    supported_outputs = OutputWorker.get_plugins()
    if supported_outputs.get('gather', None):
        del supported_outputs['gather']
    supported_outputs = list(supported_outputs.keys())

    parser = argparse.ArgumentParser()

    parser.add_argument(
        '-i',
        '--input-dir',
        type=str,
        help='Directory where run-once=gather data is'
    )

    parser.add_argument(
        '-o',
        '--outputs',
        nargs='+',
        default=['parquet'],
        choices=supported_outputs,
        type=str,
        help='Output formats to write to: parquet. Use '
        'this option multiple times for more than one output',
    )
    parser.add_argument(
        '-s',
        '--service-only',
        type=str,
        help='Only run this space separated list of services',
    )

    parser.add_argument(
        '-x',
        '--exclude-services',
        type=str,
        help='Exclude running this space separated list of services',
    )

    parser.add_argument(
        '-c',
        '--config',
        type=str, help='alternate config file'
    )

    parser.add_argument(
        '--run-once',
        type=str,
        choices=['gather', 'process', 'update'],
        help=argparse.SUPPRESS,
    )

    parser.add_argument(
        '--output-dir',
        type=str,
        default=f'{os.path.abspath(os.curdir)}/sqpoller-output',
        help=argparse.SUPPRESS,
    )

    parser.add_argument(
        '--ssh-config-file',
        type=str,
        default=None,
        help='Path to ssh config file, that you want to use'
    )

    parser.add_argument(
        '--worker-id',
        type=str,
        default='0',
        help=argparse.SUPPRESS,
    )

    userargs = parser.parse_args()

    uvloop.install()
    cfg = load_sq_config(config_file=userargs.config)
    if not cfg:
        print('Could not load config file, aborting')
        sys.exit(1)

    try:
        asyncio.run(start_worker(userargs, cfg))
    except (KeyboardInterrupt, RuntimeError):
        pass
    except Exception:  # pylint: disable=broad-except
        traceback.print_exc()

    sys.exit(0)


if __name__ == '__main__':
    worker_main()



import errno
import fcntl
import getpass
import json
import logging
import os
import re
import sys
from datetime import datetime
from enum import Enum
from importlib.util import find_spec
from ipaddress import ip_network
from itertools import groupby
from logging.handlers import RotatingFileHandler
from os import getenv
from typing import Dict, List
from tzlocal import get_localzone

import pandas as pd
import yaml
from dateparser import parse
from dateutil.relativedelta import relativedelta
from pytz import all_timezones
from suzieq.shared.exceptions import SensitiveLoadError
from suzieq.version import SUZIEQ_VERSION

logger = logging.getLogger(__name__)
MAX_MTU = 9216
# MISSING_SPEED: the interface doesn't provide a speed and I have to complain
# NO_SPEED: the interface doesn't provide a speed but I don't care
#           (for example virtual interfaces)
# MISSING_SPEED_IF_TYPES: list of interface-types that will have MISSING_SPEED
#                         if the speed is invalid.
#                         Types which are not in this list will have NO_SPEED
MISSING_SPEED = -1
NO_SPEED = 0
MISSING_SPEED_IF_TYPES = ['ethernet', 'bond', 'bond_slave']
SUPPORTED_ENGINES = ['pandas', 'rest']


class PollerTransport(str, Enum):
    """Supported poller transoport enum"""
    ssh = 'ssh'
    https = 'https'


def validate_sq_config(cfg):
    """Validate Suzieq config file

    Parameters:
    -----------
    cfg: yaml object, YAML encoding of the config file

    Returns:
    --------
    status: None if all is good or error string
    """

    if not isinstance(cfg, dict):
        return "FATAL: Invalid config file format"

    ddir = cfg.get("data-directory", None)
    if not ddir:
        return "FATAL: No data directory for output files specified"

    if not os.path.isdir(ddir):
        os.makedirs(ddir, exist_ok=True)

    if (not os.path.isdir(ddir) or not (os.access(ddir, os.R_OK | os.W_OK |
                                                  os.EX_OK))):
        if os.getenv('SQENV', None) == 'docker':
            return f'FATAL: Data directory {ddir} is not an accessible ' \
                'dir.\nIt looks like you are using docker, make sure that ' \
                'the mounted volume has the proper permissions.\nYou can ' \
                'update the permissions using the following command:\n\n' \
                'docker run --user root -v samples_parquet-db:/home/suzieq'\
                '/parquet --rm netenglabs/suzieq -c "chown -R ' \
                '1000:1000 parquet"'
        else:
            return f'FATAL: Data directory {ddir} is not an accessible dir'

    # Locate the service and schema directories
    svcdir = cfg.get('service-directory', None)
    if (not (svcdir and os.path.isdir(ddir) and
             os.access(svcdir, os.R_OK | os.W_OK | os.EX_OK))):
        sqdir = get_sq_install_dir()
        svcdir = f'{sqdir}/config'
        if os.access(svcdir, os.R_OK | os.EX_OK):
            cfg['service-directory'] = svcdir
        else:
            svcdir = None

    if not svcdir:
        return 'FATAL: No service directory found'

    schemadir = cfg.get('schema-directory', None)
    if not (schemadir and os.access(schemadir, os.R_OK | os.EX_OK)):
        schemadir = f'{svcdir}/schema'
        if os.access(schemadir, os.R_OK | os.EX_OK):
            cfg['schema-directory'] = schemadir
        else:
            schemadir = None

    if not schemadir:
        return 'FATAL: No schema directory found'

    # Move older format logging level and period to appropriate new location
    if 'poller' not in cfg:
        cfg['poller'] = {}

    for knob in ['logging-level', 'period']:
        if knob in cfg:
            cfg['poller'][knob] = cfg[knob]

    if 'rest' not in cfg:
        cfg['rest'] = {}

    for knob in ['API_KEY', 'rest_certfile', 'rest_keyfile']:
        if knob in cfg:
            cfg['rest'][knob] = cfg[knob]

    error = _load_rest_api_key(cfg)
    if error:
        return error

    # Verify timezone if present is valid
    def_tz = get_localzone().zone
    reader = cfg.get('analyzer', {})
    if reader and isinstance(reader, dict):
        usertz = reader.get('timezone', '')
        if usertz and usertz not in all_timezones:
            return f'Invalid timezone: {usertz}'
        elif not usertz:
            reader['timezone'] = def_tz
    else:
        cfg['analyzer'] = {'timezone': def_tz}

    return None


def _load_rest_api_key(cfg: Dict) -> str:
    """Loads the rest api key into the config

    Args:
        cfg (Dict): SuzieQ config

    Returns:
        str: Empty if the variable was loaded correctly,
        the error otherwise
    """
    if 'rest' in cfg and 'API_KEY' in cfg['rest']:
        api_key = cfg['rest']['API_KEY']
        if api_key == 'ask':
            # this function is called during the on_connect of the cli
            # Nubia doesn't allow the user to prompt anything at that point
            return "'ask' option is not available for REST API KEY"
        try:
            cfg['rest']['API_KEY'] = get_sensitive_data(api_key)
        except SensitiveLoadError as e:
            return f'Cannot load REST API KEY: {e}'
    return ''


def load_sq_config(validate=True, config_file=None):
    """Load (and validate) basic suzieq config"""

    # Order of looking up suzieq config:
    #   Current directory
    #   ${HOME}/.suzieq/

    cfgfile = None
    cfg = {}

    cfgfile = sq_get_config_file(config_file)

    if cfgfile:
        try:
            with open(cfgfile, "r") as f:
                cfg = yaml.safe_load(f.read())
        except Exception as e:  # pylint: disable=broad-except
            print(f'ERROR: Unable to open config file {cfgfile}: {e.args[1]}')
            sys.exit(1)

        if not cfg:
            print(f'ERROR: Empty config file {cfgfile}')
            sys.exit(1)

        if validate:
            error_str = validate_sq_config(cfg)
            if error_str:
                print(f'ERROR: Invalid config file: {cfgfile}')
                print(error_str)
                sys.exit(1)
        else:
            # also without validation, I need to load the REST API key and,
            # if necessary, raise an error
            error = _load_rest_api_key(cfg)
            if error:
                print(f'ERROR: Invalid config file: {cfgfile}')
                print(error)
                sys.exit(1)

    if not cfg:
        print("suzieq requires a configuration file either in "
              "./suzieq-cfg.yml or ~/.suzieq/suzieq-cfg.yml")
        sys.exit(1)

    return cfg


def get_sensitive_data(input_method: str, ask_message: str = '') -> str:
    """This function is used by the inventory to specify sensitive data

    The valid methods are:
        - 'plain:' (default): copy the content of input (can be omitted)
        - 'env:': get the information from an environment variable
        - 'ask': write the information on the stdin

    Args:
        input_method (str): string with info for the sensitive value
        ask_message (str): message to prompt for the 'ask' method

    Raises:
        EnvVarLoadError: environment variable not found

    Returns:
        str: sensitive data
    """
    if not input_method:
        return input_method
    sens_data = input_method
    if input_method.startswith('env:'):
        input_method = input_method.split('env:')[1].strip()
        sens_data = getenv(input_method, '')
        if not sens_data:
            raise SensitiveLoadError(
                f'No environment variable called '
                f"'{input_method}'")
    elif input_method.startswith('plain:'):
        sens_data = input_method.split("plain:")[1].strip()
    elif input_method.startswith('ask'):
        sens_data = getpass.getpass(ask_message)
    return sens_data


def sq_get_config_file(config_file):
    """Get the path to the suzieq config file"""

    if config_file:
        cfgfile = config_file
    elif os.path.exists("./suzieq-cfg.yml"):
        cfgfile = "./suzieq-cfg.yml"
    elif os.path.exists(os.getenv("HOME") + "/.suzieq/suzieq-cfg.yml"):
        cfgfile = os.getenv("HOME") + "/.suzieq/suzieq-cfg.yml"
    else:
        cfgfile = None
    return cfgfile


def get_latest_files(folder, start="", end="", view="latest") -> list:
    '''Get list of relevant parquet files from folder'''
    lsd = []

    if start:
        ssecs = pd.to_datetime(
            start, infer_datetime_format=True).timestamp() * 1000
    else:
        ssecs = 0

    if end:
        esecs = pd.to_datetime(
            end, infer_datetime_format=True).timestamp() * 1000
    else:
        esecs = 0

    ts_dirs = False
    pq_files = False

    for root, dirs, files in os.walk(folder):
        flst = None
        if dirs and dirs[0].startswith("timestamp") and not pq_files:
            flst = get_latest_ts_dirs(dirs, ssecs, esecs, view)
            ts_dirs = True
        elif files and not ts_dirs:
            flst = get_latest_pq_files(files, root, ssecs, esecs, view)
            pq_files = True

        if flst:
            lsd.append(os.path.join(root, flst[-1]))

    return lsd


def get_latest_ts_dirs(dirs, ssecs, esecs, view):
    '''Get latest timestamp directories in a folder'''
    newdirs = None

    if not ssecs and not esecs:
        dirs.sort(key=lambda x: int(x.split("=")[1]))
        newdirs = dirs
    elif ssecs and not esecs:
        newdirs = list(filter(lambda x: int(x.split("=")[1]) > ssecs, dirs))
        if not newdirs and view != "changes":
            # FInd the entry most adjacent to this one
            newdirs = list(filter(lambda x: int(
                x.split("=")[1]) < ssecs, dirs))
    elif esecs and not ssecs:
        newdirs = list(filter(lambda x: int(x.split("=")[1]) < esecs, dirs))
    else:
        newdirs = list(
            filter(
                lambda x: int(x.split("=")[1]) < esecs and int(
                    x.split("=")[1]) > ssecs,
                dirs,
            )
        )
        if not newdirs and view != "changes":
            # FInd the entry most adjacent to this one
            newdirs = list(filter(lambda x: int(
                x.split("=")[1]) < ssecs, dirs))

    return newdirs


def get_latest_pq_files(files, root, ssecs, esecs, view):
    '''Get the latest parquet files given a fileset/start & end times & view'''
    newfiles = None

    if not ssecs and not esecs:
        files.sort(key=lambda x: os.path.getctime("%s/%s" % (root, x)))
        newfiles = files
    elif ssecs and not esecs:
        newfiles = list(
            filter(lambda x: os.path.getctime(
                "%s/%s" % (root, x)) > ssecs, files)
        )
        if not newfiles and view != "changes":
            # FInd the entry most adjacent to this one
            newfiles = list(
                filter(
                    lambda x: os.path.getctime(
                        "{}/{}".format(root, x)) < ssecs, files
                )
            )
    elif esecs and not ssecs:
        newfiles = list(
            filter(lambda x: os.path.getctime(
                "%s/%s" % (root, x)) < esecs, files)
        )
    else:
        newfiles = list(
            filter(
                lambda x: os.path.getctime("%s/%s" % (root, x)) < esecs
                and os.path.getctime("%s/%s" % (root, x)) > ssecs,
                files,
            )
        )
        if not newfiles and view != "changes":
            # Find the entry most adjacent to this one
            newfiles = list(
                filter(lambda x: os.path.getctime(
                    "%s/%s" % (root, x)) < ssecs, files)
            )
    return newfiles


def calc_avg(oldval, newval):
    '''Calculate average of old and new'''

    if not oldval:
        return newval

    return float((oldval+newval)/2)


def get_timestamp_from_cisco_time(in_data, timestamp):
    """Get timestamp in ms from the Cisco-specific timestamp string
    Examples of Cisco timestamp str are P2DT14H45M16S, P1M17DT4H49M50S etc.
    """
    if not in_data.startswith('P'):
        return 0
    months = days = hours = mins = secs = 0

    if 'T' in in_data:
        day, timestr = in_data[1:].split('T')
    else:
        day = in_data[1:]
        timestr = ''

    if 'Y' in day:
        years, day = day.split('Y')
        months = int(years)*12

    if 'M' in day:
        mnt, day = day.split('M')
        months = months + int(mnt)
    if 'D' in day:
        days = int(day.split('D')[0])

    if 'H' in timestr:
        hours, timestr = timestr.split('H')
        hours = int(hours)
    if 'M' in timestr:
        mins, timestr = timestr.split('M')
        mins = int(mins)
    if 'S' in timestr:
        secs = timestr.split('S')[0]
        secs = int(secs)

    delta = relativedelta(months=months, days=days,
                          hours=hours, minutes=mins, seconds=secs)
    return int((datetime.fromtimestamp(timestamp)-delta).timestamp()*1000)


def get_timestamp_from_junos_time(in_data, timestamp: int):
    """Get timestamp in ms from the Junos-specific timestamp string
    The expected input looks like: "attributes" : {"junos:seconds" : "0"}.
    We don't check for format because we're assuming the input would be blank
    if it wasn't the right format. The input can either be a dictionary or a
    JSON string.
    """

    if not in_data:
        # Happens for logical interfaces such as gr-0/0/0
        secs = 0
    else:
        try:
            if isinstance(in_data, str):
                data = json.loads(in_data)
            else:
                data = in_data
            secs = int(data.get('junos:seconds', 0))
        except Exception:  # pylint: disable=broad-except
            logger.warning(f'Unable to convert junos secs from {in_data}')
            secs = 0

    delta = relativedelta(seconds=int(secs))
    return int((datetime.fromtimestamp(timestamp)-delta).timestamp()*1000)


def convert_macaddr_format_to_colon(macaddr: str) -> str:
    """Convert NXOS/EOS . macaddr form to standard : format, lowecase

    :param macaddr: str, the macaddr string to convert
    :returns: the converted macaddr string or all 0s string if arg not str
    :rtype: str

    """
    if isinstance(macaddr, str):
        if re.match(r'[0-9a-zA-Z]{4}.[0-9a-zA-Z]{4}.[0-9a-zA-Z]{4}', macaddr):
            return (':'.join([f'{x[:2]}:{x[2:]}'
                              for x in macaddr.split('.')])).lower()
        else:
            return macaddr.lower()

    return '00:00:00:00:00:00'


def validate_network(network: str) -> bool:
    """Validate network address

    Args:
        network: (str) the network id to validate

    Returns:
        bool: A boolean with the result of the validation

    """
    try:
        if isinstance(network, str) and '/' in network:
            ip_network(network)
            return True
        return False
    except ValueError:
        return False


def validate_macaddr(macaddr: str) -> bool:
    """Validate mac address

    Args:
        macaddr: (str) the macaddr string to validate

    Returns:
        bool: A boolean with the result of the validation

    """
    if isinstance(macaddr, str):
        if re.fullmatch(r'([0-9a-fA-F]{4}.){2}[0-9a-fA-F]{4}', macaddr) or \
           re.fullmatch(r'([0-9a-fA-F]{2}:){5}[0-9a-fA-F]{2}', macaddr):
            return True

    return False


def convert_rangestring_to_list(rangestr: str) -> list:
    """Convert a range list such as '1, 2-5, 10, 12-20' to list
    """

    tmplst = []
    if not isinstance(rangestr, str):
        return tmplst

    try:
        for x in rangestr.split(','):
            x = x.strip().split('-')
            if x[0]:
                if len(x) == 2:
                    intrange = list(range(int(x[0]), int(x[1])+1))
                    tmplst.extend(intrange)
                else:
                    tmplst.append(int(x[0]))
    except Exception:  # pylint: disable=broad-except
        logger.error(f"Range string parsing failed for {rangestr}")
        return []
    return tmplst


def convert_numlist_to_ranges(numList: List[int]) -> str:
    """Convert a given list of numbers into a range string

    Args:
        numList (List[int]): unsorted/sorted list of integers

    Returns:
        str: Range string such as '1-5, 10, 12-20'
    """
    result = ''
    for _, b in groupby(enumerate(sorted(numList)),
                        lambda pair: pair[1] - pair[0]):
        b = list(b)
        if len(b) > 1:
            result += f'{b[0][1]}-{b[-1][1]}, '
        else:
            result += f'{b[0][1]}, '

    return result[:-2]


def build_query_str(skip_fields: list, schema, ignore_regex=True,
                    **kwargs) -> str:
    """Build a pandas query string given key/val pairs
    """
    query_str = ''
    prefix = ''

    def _build_query_str(fld, val, fldtype) -> List[str]:
        """Builds the string from the provided user input

        Besides handling operators and regexp, the function returns what
        the subsequent match within this list needs to be. If the ! operator
        is used, we have to switch to AND, else we can be at OR. For
        example. it makes no sense to say !leaf01 OR !leaf02 as this will
        include both leaf01 and leaf02.
        """

        cond = 'or'
        num_type = ["long", "float", "int"]
        if ((fldtype in num_type) and not
                isinstance(val, str)):
            result = f'{fld} == {val}'

        elif val.startswith('!'):
            val = val[1:]
            cond = 'and'
            if fldtype in num_type:
                result = f'{fld} != {val}'
            else:
                result = f'{fld} != "{val}"'
        elif val.startswith(('<', '>')):
            result = val
        elif val.startswith('~'):
            val = val[1:]
            if val.startswith('!'):
                val = val[1:]
                cond = 'and'
                result = f'~{fld}.str.match("{val}")'
            else:
                result = f'{fld}.str.match("{val}")'
        else:
            if fldtype in num_type:
                result = f'{fld} == {val}'
            else:
                result = f'{fld} == "{val}"'

        return result, cond

    for f, v in kwargs.items():
        if not v or f in skip_fields or f in ["groupby"]:
            continue

        stype = schema.field(f).get('type', 'string')
        if isinstance(v, list) and len(v):
            notlist = [x for x in v
                       if x.startswith('!') or x.startswith('~!')]
            if notlist != v:
                newv = [x for x in v
                        if not x.startswith('!') and not x.startswith('~!')]
            else:
                newv = v
            subq = ''
            subcond = ''
            if ignore_regex and [x for x in newv
                                 if isinstance(x, str) and
                                 x.startswith('~')]:
                continue

            for elem in newv:
                intq, nextcond = _build_query_str(f, elem, stype)
                subq += f'{subcond} {intq} '
                subcond = nextcond
            query_str += '{} ({})'.format(prefix, subq)
            prefix = "and"
        else:
            intq, nextcond = _build_query_str(f, v, stype)
            query_str += f'{prefix} {intq} '
            prefix = nextcond

    return query_str


def poller_log_params(cfg: dict, is_controller=False, worker_id=0) -> tuple:
    """Get the log file, level and size for the given program from config
    It gets the base file name of the configuration file and appends a prefix
    which depends on the component of the poller

    Args:
        cfg (dict): The config dictionary
        is_controller (bool, optional): If the component is the controller.
            Defaults to False.
        worker_id (int, optional): The poller worker id. Defaults to 0.

    Returns:
        tuple: [description]
    """
    def_logfile = '/tmp/sq-poller.log'
    logfile, loglevel, logsize, log_stdout = get_log_params(
        'poller', cfg, def_logfile)
    file_name = logfile.split('.log')[0]
    if is_controller:
        file_name += '-controller.log'
    else:
        file_name += f'-{worker_id}.log'
    return file_name, loglevel, logsize, log_stdout


def get_log_params(prog: str, cfg: dict, def_logfile: str) -> tuple:
    """Get the log file, level and size for the given program from config

    The logfile is supposed to be defined by a variable called logfile
    within the hierarchy of the config dictionary. Thus, the poller log file
    will be {'poller': {'logfile': '/tmp/sq-poller.log'}}, for example.

    :param prog: str, The name of the program. Valid values are poller,
                      coaelscer, and rest.
    :param cfg: dict, The config dictionary
    :param def_logfile: str, The default log file to return
    :returns: log file name, log level, log size, and
              True/False for logging to stdout
    :rtype: str, str and int

    """
    if cfg:
        logfile = cfg.get(prog, {}).get('logfile', def_logfile)
        loglevel = cfg.get(prog, {}).get('logging-level', 'WARNING')
        logsize = cfg.get(prog, {}).get('logsize', 10000000)
        log_stdout = cfg.get(prog, {}).get('log-stdout', False)
    else:
        logfile = def_logfile
        loglevel = 'WARNING'
        logsize = 10000000
        log_stdout = False

    return logfile, loglevel, logsize, log_stdout


def init_logger(logname: str,
                logfile: str,
                loglevel: str = 'WARNING',
                logsize: int = 10000000,
                use_stdout: bool = False) -> logging.Logger:
    """Initialize the logger

    :param logname: str, the name of the app that's logging
    :param logfile: str, the log file to use
    :param loglevel: str, the default log level to set the logger to
    :param use_stdout: str, log to stdout instead of or in addition to file

    """

    fh = sh = None
    # this needs to be suzieq.poller, so that it is the root of all the
    # other pollers
    log = logging.getLogger(logname)
    log.setLevel(loglevel.upper())
    if logfile:
        fh = RotatingFileHandler(logfile, maxBytes=logsize, backupCount=2)
    if use_stdout:
        sh = logging.StreamHandler(sys.stdout)

    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s " "- %(message)s"
    )
    if fh:
        fh.setFormatter(formatter)
    if sh:
        sh.setFormatter(formatter)

    # set root logger level, so that we set asyncssh log level
    #  asynchssh sets it's level to the root level
    root = logging.getLogger()
    root.setLevel(loglevel.upper())
    if fh:
        root.addHandler(fh)
    if sh:
        root.addHandler(sh)

    log.warning(f"log level {logging.getLevelName(log.level)}")

    return log


def known_devtypes() -> list:
    """Returns the list of known dev types"""
    return(['cumulus', 'eos', 'iosxe', 'iosxr', 'ios', 'junos-mx', 'junos-qfx',
            'junos-qfx10k', 'junos-ex', 'junos-es', 'linux', 'nxos', 'sonic',
            'panos'])


def humanize_timestamp(field: pd.Series, tz=None) -> pd.Series:
    '''Convert the UTC timestamp in Dataframe to local time.
    Use of pd.to_datetime will not work as it converts the timestamp
    to UTC. If the timestamp is already in UTC format, we get busted time.
    '''
    if field.empty:
        return field

    if pd.core.dtypes.common.is_datetime_or_timedelta_dtype(field):
        return field
    if pd.core.dtypes.common.is_datetime64_any_dtype(field):
        return field
    tz = tz or get_localzone().zone
    return field.apply(lambda x: datetime.utcfromtimestamp((int(x)/1000))) \
                .dt.tz_localize('UTC').dt.tz_convert(tz)


def expand_nxos_ifname(ifname: str) -> str:
    '''Expand shortned ifnames in NXOS to their full values, if required'''
    if not ifname:
        return ''
    if ifname.startswith('Eth') and 'Ether' not in ifname:
        return ifname.replace('Eth', 'Ethernet')
    elif ifname.startswith('Po') and 'port' not in ifname:
        return ifname.replace('Po', 'port-channel')
    return ifname


def expand_eos_ifname(ifname: str) -> str:
    '''Expand shortned ifnames in EOS to their full values, if required'''
    if not ifname:
        return ''
    if ifname.startswith('Eth') and 'Ether' not in ifname:
        return ifname.replace('Eth', 'Ethernet')
    elif ifname.startswith('Po') and 'Port' not in ifname:
        return ifname.replace('Po', 'Port-Channel')
    elif ifname.startswith('Vx') and 'Vxlan' not in ifname:
        return ifname.replace('Vx', 'Vxlan')
    return ifname


def ensure_single_instance(filename: str, block: bool = False) -> int:
    """Check there's only a single active instance of a process using lockfile

    It optionally can block waiting for the resource the become available.

    Use a pid file with advisory file locking to assure this.

    :returns: fd if lock was successful or 0
    :rtype: int

    """
    basedir = os.path.dirname(filename)
    if not os.path.exists(basedir):
        # Permission error or any other error will abort
        os.makedirs(basedir, exist_ok=True)

    fd = os.open(filename, os.O_RDWR | os.O_CREAT, 0o600)
    if fd:
        try:
            if block:
                fcntl.flock(fd, fcntl.LOCK_EX)
            else:
                fcntl.flock(fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
            os.truncate(fd, 0)
            os.write(fd, bytes(str(os.getpid()), 'utf-8'))
        except OSError:
            if OSError.errno == errno.EBUSY:
                # Return the PID of the process thats locked the file
                bpid = os.read(fd, 10)
                os.close(fd)
                try:
                    fd = -int(bpid)
                except ValueError:
                    fd = 0
            else:
                os.close(fd)
                fd = 0

    return fd


def expand_ios_ifname(ifname: str) -> str:
    """Get expanded interface name for IOSXR/XE given its short form

    :param ifname: str, short form of IOSXR interface name
    :returns: Expanded version of short form interface name
    :rtype: str
    """

    ifmap = {'BE': 'Bundle-Ether',
             'BV': 'BVI',
             'Fi': 'FiftyGigE',
             'Fo': 'FortyGigE',
             'FH': 'FourHundredGigE',
             'Gi': 'GigabitEthernet',
             'Gig': 'GigabitEthernet',
             'Hu': 'HundredGigE',
             'Lo': 'Loopback',
             'Mg': 'MgmtEth',
             'Nu': 'Null',
             'Po': 'Port-channel',
             'TE': 'TenGigE',
             'Te': 'TenGigabitEthernet',
             'Ten': 'TenGigabitEthernet',
             'TF': 'TwentyFiveGigE',
             'TH': 'TwoHundredGigE',
             'tsec': 'tunnel-ipsec',
             'tmte': 'tunnel-mte',
             'tt': 'tunnel-te',
             'tp': 'tunnel-tp',
             'Vl': 'Vlan',
             'CPU': 'cpu',
             }
    pfx = re.match(r'[a-zA-Z]+', ifname)
    if pfx:
        pfxstr = pfx.group(0)
        if pfxstr in ifmap:
            return ifname.replace(pfxstr, ifmap[pfxstr])

    return ifname


def get_sq_install_dir() -> str:
    '''Return the absolute path of the suzieq installation dir'''
    spec = find_spec('suzieq')
    if spec:
        return os.path.dirname(spec.loader.path)
    else:
        return os.path.abspath('./')


def get_sleep_time(period: str) -> int:
    """Returns the duration in seconds to sleep given a period

    Checking if the period format matches a specified format MUST be
    done by the caller.

    :param period: str, the period of form <value><unit>, '15m', '1h' etc
    :returns: duration to sleep in seconds
    :rtype: int
    """
    _, unit, _ = re.split(r'(\D)', period)
    now = datetime.now()
    nextrun = parse(period, settings={'PREFER_DATES_FROM': 'future'})
    if unit == 'm':
        nextrun = nextrun.replace(second=0)
    elif unit == 'h':
        nextrun = nextrun.replace(minute=0, second=0)
    else:
        nextrun = nextrun.replace(hour=0, minute=0, second=0)

    return (nextrun-now).seconds


def convert_asndot_to_asn(asn: str) -> int:
    """Convert BGP ASN into asdot format

    Convert BGP ASN if a single integer to asdot(<asn_hi>.<asn_lo>)
    format. If input ASN is in asdot format already, it returns it as is.
    If input ASN is < 65535, returns it as is.

    Args:
        asn: ASN to convert

    Returns:
        BGP ASN as 32b integer
    """

    if isinstance(asn, int) or '.' not in asn:
        return asn

    s_asn = asn.split('.')
    return int(s_asn[0])*65536+int(s_asn[1])


def print_version():
    '''Print the suzieq version and return'''
    print(SUZIEQ_VERSION)



'''
SuzieQ Page interface.
Any page built inside the SuzieQ GUI has to adhere to this interface
'''
from abc import ABC, abstractmethod
from dataclasses import asdict

import streamlit as st
from suzieq.shared.sq_plugin import SqPlugin


class SqGuiPage(ABC, SqPlugin):
    '''A SuzieQ GUI Page Interface class
    Every page has access to the config and context via the state variable
    passed in building the page
    '''

    _state = {}  # Set it to a dataclass specific to the page
    _title: str = None  # Initialized by each page
    _first_time: bool = True
    _URL_PARAMS_BLACKLIST = ['get_clicked', 'uniq_clicked', 'tables_obj']
    _LIST_URL_PARAMS = ['columns']

    @property
    @abstractmethod
    def add_to_menu(self) -> bool:
        '''Should this page be added to the menu?'''
        return NotImplementedError

    @property
    def title(self) -> str:
        '''Page title

        Mandatory function if you want to display a page
        '''
        return self._title

    @abstractmethod
    def build(self) -> None:
        '''The main page building routine'''
        # The basic template is:
        # 1. Create sidebar (all user input is provided via sidebar)
        # 2. Create layout
        # 3. Perform appropriate computations
        # 4. Display the info using the layout
        # 5. Sync widget state if necessary (status page does nothing for this)
        # 6. Save params as a URL to share page with others
        raise NotImplementedError

    def _get_state_from_url(self):
        url_params = st.experimental_get_query_params()
        page = url_params.pop('page', '')

        if self._title in page:

            if url_params and not all(not x for x in url_params.values()):
                for key, val in url_params.items():
                    # Do not add the params that are in the blacklist
                    if key in self._URL_PARAMS_BLACKLIST:
                        continue
                    if (isinstance(val, list) and
                            key not in self._LIST_URL_PARAMS):
                        val = val[0].strip()
                    if not val:
                        continue
                    if hasattr(self._state, key):
                        if isinstance(getattr(self._state, key), bool):
                            val = val == 'True'
                        setattr(self._state, key, val)

    @abstractmethod
    def _create_sidebar(self) -> None:
        '''Draw the sidebar for this page'''
        raise NotImplementedError

    @abstractmethod
    def _create_layout(self) -> None:
        '''Create the page layout

        By creating a layout first, you avoid screen flicker and such
        when rendering the page first and then loading the data. This
        is specifically the case with pages like Xplore.
        '''
        raise NotImplementedError

    @abstractmethod
    def _render(self, layout: dict) -> None:
        '''Render the page with all the computations'''
        raise NotImplementedError

    @abstractmethod
    def _sync_state(self) -> None:
        '''Save the widget states'''
        raise NotImplementedError

    def _save_page_url(self) -> None:
        '''Save the page params as an URL for sharing'''
        state = asdict(self._state)
        for param in self._URL_PARAMS_BLACKLIST:
            state.pop(param, None)
        st.experimental_set_query_params(**state)



"""This module contains the class to import device credentials using files
"""
# pylint: disable=no-name-in-module
# pylint: disable=no-self-argument
import logging
from pathlib import Path
from typing import Dict, List, Optional, Union
from pydantic import BaseModel, Field, validator

import yaml
from suzieq.poller.controller.credential_loader.base_credential_loader import \
    CredentialLoader, CredentialLoaderModel, check_credentials
from suzieq.shared.exceptions import InventorySourceError

logger = logging.getLogger(__name__)


class CredFileEntryModel(BaseModel):
    """Model to validate entries in credential file
    """
    hostname: Optional[str]
    address: Optional[str]
    username: Optional[str]
    password: Optional[str]
    keyfile: Optional[str]
    key_passphrase: Optional[str] = Field(alias='key-passphrase')
    enable_password: Optional[str] = Field(alias='enable-password')

    class Config:
        """pydantic configuration
        """
        extra = 'forbid'


class CredFileNamespaceModel(BaseModel):
    """Model to validate the content of the credential file
    """
    namespace: str
    devices: List[CredFileEntryModel]

    class Config:
        """pydantic configuration
        """
        extra = 'forbid'


class CredFileModel(CredentialLoaderModel):
    """Model for credential file validation
    """
    credentials: Union[str, List[CredFileNamespaceModel]] = Field(alias='path')

    @validator('credentials')
    def validate_credentials(cls, cred):
        """validate the credentials
        """
        if isinstance(cred, str):
            dev_cred_file = Path(cred)
            if not dev_cred_file.is_file():
                raise ValueError(
                    f'The credential file {cred} does not exists')

            try:
                with open(dev_cred_file, 'r') as f:
                    credentials = yaml.safe_load(f.read())
            except yaml.YAMLError:
                raise ValueError(
                    'The credential file is not a valid yaml file'
                )

            if not credentials:
                raise ValueError(
                    'The credential file is empty'
                )

            if not isinstance(credentials, list):
                raise ValueError(
                    'The credentials file must contain all device '
                    'credential divided in namespaces'
                )
        elif isinstance(cred, list):
            credentials = cred
        valid_creds = []
        for c in credentials:
            if not isinstance(c, dict):
                raise ValueError(
                    'Namespaces credentials must be a dictionary')
            valid_creds.append(CredFileNamespaceModel(**c))
        return valid_creds


class CredFile(CredentialLoader):
    """Reads devices credentials from a file and write them on the inventory
    """

    @classmethod
    def get_data_model(cls):
        return CredFileModel

    def init(self, init_data: dict):
        if not self._validate:
            # fix pydantic aliases
            init_data['credentials'] = []
            for ns_cred in init_data.pop('path', []):
                init_data['credentials'].append(
                    CredFileNamespaceModel(**ns_cred)
                )

        super().init(init_data)
        if not self._data.credentials:
            raise InventorySourceError(f'{self.name} empty credentials')

    def load(self, inventory: Dict):

        for ns_credentials in self._data.credentials:
            namespace = ns_credentials.namespace
            if not namespace:
                raise InventorySourceError(
                    f'{self.name} All namespaces must have a name')

            ns_nodes = ns_credentials.devices
            if not ns_nodes:
                logger.warning(
                    f'{self.name} No devices in {namespace} namespace')
                continue

            for ns_node in ns_nodes:
                node_info = ns_node.dict(by_alias=True)
                if node_info.get('hostname'):
                    node_id = node_info['hostname']
                    node_key = 'hostname'
                elif node_info.get('address'):
                    node_id = node_info['address']
                    node_key = 'address'
                else:
                    raise InventorySourceError(
                        f'{self.name} Nodes must have a hostname or '
                        'address')

                node = [x for x in inventory.values()
                        if x.get(node_key) == node_id]
                if not node:
                    logger.warning(
                        f'{self.name} Unknown node called {node_id}')
                    continue

                node = node[0]
                if namespace != node.get('namespace', ''):
                    raise InventorySourceError(
                        f'The device {node_id} does not belong the namespace '
                        f'{namespace}'
                    )
                # rename 'keyfile' into 'ssh_keyfile'
                node_info['ssh_keyfile'] = node_info.pop('keyfile', None)

                # rename 'key-passphrase' into 'passphrase'
                node_info['passphrase'] = node_info.pop(
                        'key-passphrase', None)

                # rename 'enable-password' into 'enable_password'
                node_info['enable_password'] = node_info.pop(
                    'enable-password', None)

                node_cred = node_info.copy()

                node_cred.pop('address')
                node_cred.pop('hostname')

                fields = ['username', 'passphrase', 'ssh_keyfile', 'password']
                multi_defined = []
                for f in fields:
                    if node.get(f) and node_cred.get(f):
                        multi_defined.append(f)
                    elif node.get(f):
                        node_cred[f] = node[f]

                if multi_defined:
                    raise InventorySourceError(
                        f"{self.name} the node {node.get('address')} has the "
                        "following strings defined in multiple places "
                        f"{multi_defined}")

                self.write_credentials(node, node_cred)

        check_credentials(inventory)



# pylint: disable=no-name-in-module

import re
from typing import Dict, List, Optional
from pydantic import BaseModel, Field, validator

from suzieq.shared.utils import PollerTransport


class InventoryModel(BaseModel):
    """Model for the inventory validation
    """
    sources: List[Dict] = Field(min_items=1)
    namespaces: List[Dict] = Field(min_items=1)
    auths: Optional[List[Dict]]
    devices: Optional[List[Dict]]

    class Config:
        """pydantic configuration
        """
        extra = 'forbid'


class DeviceModel(BaseModel):
    """Device model validation
    """
    name: str
    jump_host: Optional[str] = Field(alias='jump-host')
    jump_host_key_file: Optional[str] = Field(alias='jump-host-key-file')
    ignore_known_hosts: Optional[bool] = Field(
        alias='ignore-known-hosts', default=False)
    slow_host: Optional[bool] = Field(alias='slow-host', default=False)
    transport: Optional[PollerTransport]
    port: Optional[str]
    devtype: Optional[str]

    @validator('jump_host')
    def jump_host_validator(cls, value):
        jump_host = value[2:] if value.startswith('//') else value
        pattern = '([a-zA-Z0-9-.]{1,253})@([a-zA-Z0-9-.]{1,253})'
        assert re.fullmatch(pattern, jump_host), \
               'format username@jumphost required'
        return value

    class Config:
        """pydantic configuration
        """
        extra = 'forbid'


class NamespaceModel(BaseModel):
    """Namespace model validation
    """
    name: str
    source: str
    device: Optional[str]
    auth: Optional[str]

    class Config:
        """pydantic configuration
        """
        extra = 'forbid'



"""
This module implements all the logic related to initialization of the services
"""
import asyncio
import logging
import os
from collections import defaultdict
from pathlib import Path
from time import struct_time
from typing import Callable, Dict, List, Union

import textfsm
import yaml
from genericpath import isfile
from suzieq.db.base_db import SqDB

from suzieq.poller.worker.services.service import Service
from suzieq.shared.exceptions import SqPollerConfError
from suzieq.shared.schema import Schema, SchemaForTable

logger = logging.getLogger(__name__)


class ServiceManager:
    """Service manager is the component in charge of the
    Services setup and initialization.
    """

    def __init__(self,
                 add_task_fn: Callable,
                 service_directory: str,
                 schema_dir: str,
                 output_queue: asyncio.Queue,
                 run_mode: str,
                 cfg: Dict,
                 default_interval: int = 15,
                 **kwargs) -> None:
        """Instantiate an instance of the ServiceManager class

        Args:
            add_task_fn (Callable): the function to call to schedule
                a task in the poller.
            service_directory (str): the directory where services are
                described.
            schema_dir (str): the directory containing the schema of the
                tables of the services.
            output_queue (asyncio.Queue): it is the queue where the services
                are supposed to write to make a writer persist the output of
                a command.
            run_mode (str): a string representing the running policy:
                - gather (run once): gather the output without processing
                - processing (run once): gather the output and process it
                - forever: periodically poll the nodes.
            default_interval (int): the default time execution interval.
        """

        self._services = []
        self._running_svcs = []
        self.add_task_fn = add_task_fn
        self.output_queue = output_queue
        self.default_interval = default_interval
        self.run_mode = run_mode
        self.cfg = cfg
        self.outputs = kwargs.pop('outputs', [])

        # Set and validate service and schema directories
        if not os.path.isdir(service_directory):
            raise SqPollerConfError(
                f"Service directory {service_directory} is not a directory"
            )
        self.service_directory = service_directory
        self.schema_dir = schema_dir
        if not self.schema_dir:
            self.schema_dir = '{}/{}'.format(service_directory, 'schema')
        if not os.path.isdir(self.schema_dir):
            raise SqPollerConfError(
                f"Service directory {self.schema_dir} is not a directory"
            )

        # Build the list of services to execute
        service_only = kwargs.pop('service_only', '')
        exclude_services = kwargs.pop('exclude_services', '')
        self._svcs_list = self._get_service_list(service_only,
                                                 exclude_services)

    @property
    def services(self) -> List[Service]:
        """Get the list of Service objects

        Returns:
            List[Service]: list of the instantiated Service
                objects, returns an empty list if the init_services()
                has not been called.
        """
        return self._services

    @property
    def svcs_list(self) -> List[str]:
        """Get the list of the name of the services

        Returns:
            List[str]: a list of string containing the name of the services
                selected for execution
        """
        return self._svcs_list

    @property
    def running_services(self) -> List[Callable]:
        """Get the list of running services

        Returns:
            List[Callable]: the list of coroutines of the running
                services.
        """
        return self._running_svcs

    @staticmethod
    def get_service_list(service_only: str,
                         exclude_services: str,
                         service_directory: struct_time) -> List[str]:
        """Constuct the list of the name of the services to be started
        in the poller, according to the arguments passed by the user.
        Returns a InvalidAttribute exception if any of the passed service
        is invalid.

        Args:
            service_only (str): take only the services in the space-separated
                list
            exclude_service (List[str]): remove the given services from the
                space-separated service list
            service_directory (str): the directory where services are defined

        Raises:
            SqPollerConfError: raised in case of wrong service name in
            'include only' or exclude list

        Returns:
            List[str]: the list of services to executed in the poller
        """

        BLACKLIST_SERVICES = ['ifCounters', 'topmem', 'topcpu']

        if not os.path.isdir(service_directory):
            raise SqPollerConfError(
                'The service directory provided is not a directory'
            )

        svcs = list(Path(service_directory).glob('*.yml'))
        allsvcs = [os.path.basename(x).split('.')[0] for x in svcs
                   if x not in BLACKLIST_SERVICES]
        svclist = None

        if service_only:
            svclist = service_only.split()

            # Check if all the given services are valid
            notvalid = [s for s in svclist if s not in allsvcs]
            if notvalid:
                raise SqPollerConfError(f'Invalid svcs specified: {notvalid}. '
                                        f'Should have been one of {allsvcs}')
        else:
            svclist = allsvcs

        if exclude_services:
            excluded_services = exclude_services.split()
            # Check if all the excluded services are valid
            notvalid = [e for e in excluded_services if e not in allsvcs]
            if notvalid:
                raise SqPollerConfError(f'Services {notvalid} excluded, but '
                                        'they are not valid.')
            svclist = list(filter(lambda x: x not in excluded_services,
                                  svclist))

        if not svclist:
            raise SqPollerConfError('The list of services to execute is empty')
        return svclist

    async def init_services(self) -> List[Service]:
        """Instantiate Service objects and prepare
        them for running. This function should be called before
        scheduling the service for running.

        Returns:
            List[Service]: the list of the initialized service instances
        """
        services = []
        schemas = defaultdict(dict)
        svc_classes = Service.get_plugins()

        schemas = Schema(self.schema_dir)
        if schemas:
            poller_schema = schemas.get_arrow_schema('sqPoller')
            poller_schema_version = SchemaForTable('sqPoller', schemas).version

        db_access = self._get_db_access(self.cfg)

        # Read the available services and iterate over them, discarding
        # the ones we do not need to instantiate
        svc_desc_files = Path(self.service_directory).glob('*.yml')

        for filename in svc_desc_files:
            with open(filename, 'r') as f:
                svc_def = yaml.safe_load(f.read())

            if not svc_def:
                logger.warning(f'Skip empty service file: {filename}')
                continue

            if svc_def.get('service') not in self.svcs_list:
                logger.warning(
                    f"Ignoring unspecified service {svc_def.get('service')}"
                )
                continue

            if 'service' not in svc_def or 'apply' not in svc_def:
                logger.error(
                    'Ignoring invalid service file definition.'
                    f"'service' and 'apply' keywords: {filename}"
                )
                continue

            period = svc_def.get('period', self.default_interval)
            for nos, cmds_desc in svc_def['apply'].items():

                # Check if the the current nos copies from another
                if isinstance(cmds_desc, dict) and 'copy' in cmds_desc:
                    newval = svc_def['apply'].get(cmds_desc['copy'], None)
                    if not newval:
                        logger.error(
                            f"No device type {cmds_desc['copy']} to copy from,"
                            f"for {nos} for service {svc_def['service']}"
                        )
                        return
                    cmds_desc = newval

                # Update the command description adding the
                # specification for the output parsing
                if isinstance(cmds_desc, list):
                    for subele in cmds_desc:
                        self._parse_nos_version(filename, svc_def, nos, subele)
                else:
                    self._parse_nos_version(filename, svc_def, nos, cmds_desc)

            try:
                schema = SchemaForTable(svc_def['service'], schema=schemas)
            except Exception:  # pylint: disable=broad-except
                logger.error(f"No matching schema for {svc_def['service']}")
                continue

            if schema.type == 'derivedRecord':
                # These are not real services and so ignore them
                continue

            # Valid service definition, add it to list
            # if the service has not a dedicated class, we will use the
            # default implementation
            class_to_use = svc_classes.get(svc_def['service'], Service)
            service = class_to_use(
                svc_def['service'],
                svc_def['apply'],
                period,
                svc_def.get('type', 'state'),
                svc_def.get('keys', []),
                svc_def.get('ignore-fields', []),
                schema,
                self.output_queue,
                db_access,
                self.run_mode
            )
            service.poller_schema = poller_schema
            service.poller_schema_version = poller_schema_version
            logger.info(f'Service {service.name} added')
            services.append(service)

        # Once done set the service list and return its content
        self._services = services
        return self._services

    async def schedule_services_run(self):
        """Schedule the services tasks in the poller, so that they can
        start sending command queries to the nodes.

        This function should be called only once, if called more
        than once, it won't have any effect.
        """
        self._running_svcs = [svc.run() for svc in self.services]
        await self.add_task_fn(self._running_svcs)

    async def set_nodes(self, node_callq: Dict):
        """Set/Update the list of nodes to poll

        Args:
            node_callq (Dict): a dictionary having, for each
                'namespace.hostname' the node hostname and the function
                allowing to calla query on the node.
        """
        for svc in self._services:
            await svc.set_nodes(node_callq)

    def _get_db_access(self, cfg) -> SqDB:
        """Return the SqDB to use to access to the state of the previous
        polls

        Raises:
            DBNotFoundError: raised if not plugin for the given outputs is
                found

        Returns:
            SqDB: the SqBB object to use for data access
        """
        if not self.outputs:
            return None
        # Remove gather from the outputs, since with it we only write in files
        candidate_out = [o for o in self.outputs if o != 'gather']
        if candidate_out:
            dbs = SqDB.get_plugins()
            # Get only the first out as source
            outdb = candidate_out[0]
            if outdb not in dbs:
                raise SqPollerConfError(f'{outdb} database not found')
            # Init the SqDB object
            return dbs[outdb](cfg, logger)
        return None

    def _get_service_list(self,
                          service_only: str,
                          exclude_services: str) -> List[str]:
        """Call the get_service_directory static method to constuct the list
        of the name of the services to be started.

        Args:
            service_only (str): take only the services in the space-separated
                list
            exclude_service (List[str]): remove the given services from the
                space-separated service list

        Raises:
            SqPollerConfError: raised in case of wrong service name in
            'include only' or exclude list

        Returns:
            List[str]: the list of services to executed in the poller
        """
        return self.get_service_list(service_only,
                                     exclude_services,
                                     self.service_directory)

    # pylint: disable=unused-argument
    def _parse_nos_version(self,
                           filename: str,
                           svc_def: Dict,
                           nos: str,
                           cmds_desc: Union[Dict, List]):
        """Given a command description check whether initialize the textfsm
        finite state machine for the output parsing, if needed.

        Args:
            filename (str): the name of the file describing the service
            svc_def (Dict): the definition of the service
            nos (str): the nos version the commands are related to
            cmds_desc (Union[Dict, List]): the description fo the commands to
                execute.
        """
        # pylint: disable=too-many-boolean-expressions
        if ('command' not in cmds_desc) or (
                (
                    isinstance(cmds_desc['command'], list)
                    and not all('textfsm' in x or 'normalize' in x
                                for x in cmds_desc['command'])
                )
                or (
                    not isinstance(cmds_desc['command'], list)
                    and ('normalize' not in cmds_desc
                         and 'textfsm' not in cmds_desc)
                )
        ):
            logger.error(
                'Ignoring invalid service file '
                "definition. Need both 'command' and "
                f"'normalize/textfsm' keywords: {filename}, {cmds_desc}"
            )
            return

        if 'textfsm' in cmds_desc:
            # We may have already visited this element and parsed
            # the textfsm file. Check for this and in this case return
            if cmds_desc['textfsm'] and isinstance(
                cmds_desc['textfsm'], textfsm.TextFSM
            ) or (cmds_desc['textfsm'] is None):
                return

            # Read the file and create the textfsm in the cmd description
            tfsm_file = self.service_directory + '/' + cmds_desc['textfsm']
            if not isfile(tfsm_file):
                logger.error(
                    f'Textfsm file {tfsm_file} not found. Ignoring service'
                )
                return

            with open(tfsm_file, 'r') as f:
                tfsm_template = textfsm.TextFSM(f)
                cmds_desc['textfsm'] = tfsm_template
        if isinstance(cmds_desc['command'], list):
            # If the command description is a list of commands
            # we need to iterate over them
            for subelem in cmds_desc['command']:
                if 'textfsm' in subelem:
                    # Check if the elemnt has been already visited
                    if subelem['textfsm'] and isinstance(
                        subelem['textfsm'], textfsm.TextFSM
                    ):
                        continue

                    tfsm_file = (f"{self.service_directory}/"
                                 f"{subelem['textfsm']}")
                    if not isfile(tfsm_file):
                        logger.error(
                            f'Textfsm file {tfsm_file} not found. Ignoring'
                            ' service'
                        )
                        continue

                    with open(tfsm_file, 'r') as f:
                        try:
                            tfsm_template = textfsm.TextFSM(f)
                            subelem['textfsm'] = tfsm_template
                        except Exception:  # pylint: disable=broad-except
                            logger.exception(
                                'Unable to load TextFSM file '
                                f'{tfsm_file} for service '
                                f"{svc_def['service']}")
                            continue



from nubia import command
from suzieq.cli.nubia_patch import argument

from suzieq.cli.sqcmds.command import SqCommand
from suzieq.sqobjects.path import PathObj


@command("path", help="build and act on path data")
@argument("src", description="Source IP address, in quotes")
@argument("dest", description="Destination IP address, in quotes")
@argument("vrf", description="VRF to trace path in")
class PathCmd(SqCommand):
    """Path trace information including overlay and underlay"""

    def __init__(
        self,
        engine: str = "",
        hostname: str = "",
        start_time: str = "",
        end_time: str = "",
        view: str = "",
        namespace: str = "",
        format: str = "",  # pylint: disable=redefined-builtin
        columns: str = "default",
        src: str = "",
        dest: str = "",
        vrf: str = "",
    ) -> None:
        super().__init__(
            engine=engine,
            hostname=hostname,
            start_time=start_time,
            end_time=end_time,
            view=view,
            namespace=namespace,
            columns=columns,
            format=format,
            sqobj=PathObj
        )
        self.lvars = {
            'src': src,
            'dest': dest,
            'vrf': vrf
        }



#!/usr/bin/env python3

import asyncio
import re
import json
import logging
from pathlib import Path
from http import HTTPStatus
from datetime import datetime, timezone
import aiofiles

from suzieq.poller.worker.services.service import RsltToken


class FileNode:
    '''Class using run-once=gather snapshot output as mock device input'''

    # pylint: disable=attribute-defined-outside-init
    async def initialize(self, datadir: str):
        '''Load the data to be read from files'''
        self.logger = logging.getLogger(__name__)
        self.data = {}          # The dict of command to output entries
        self.hostname = '_filedata'  # Need this for the services

        self._service_queue = asyncio.Queue()

        await self._load_data(datadir)

    async def _load_data(self, datadir: str):
        """This routine walks the dir loading the data from the files there"""
        p = Path(datadir)
        tasks = []
        for file in p.glob('*.output'):
            tasks.append(self._load_single_file_data(file))

        if not tasks:
            self.logger.error(f'No files to load found in {datadir}')
            return []

        await asyncio.gather(*tasks)

    async def _load_single_file_data(self, file: str):
        """Load gathered data from specified file
        Each entry loaded from the file is a list entry with the format:
        [{'status':...
         'timestamp': ...
         'cmd': ....
         'devtype': ...
         'namespace': ...
         'hostname': ...
         'address': ...
         'version': ...
         'data': ...}]

        If the service consisted of multiple commands, then the format is a
        list, where each entry in the list is of the above format.

        We return a dictionary keyed on the command of the format:
        {cmd1: {curpos: 0, 'entry': <list of entries>},
         cmd2: {curpos: 0, 'entry': <list of entries>,
         ...
        } where each entry in the list is of the individual entry of the input
        format. The basic idea is that when input is requested for a given cmd,
        we return the entry at curpos and increment curpos. When curpos
        exceeds the length of the entries, we reset it back to 0.
        """
        async with aiofiles.open(file, 'r') as f:
            data = await f.read()

        required_keys = ['status', 'timestamp', 'cmd', 'devtype', 'namespace',
                         'hostname', 'address', 'version', 'data']
        entries = re.split(r'\]\n*\[\n', data)
        entlen = len(entries)

        for i, elem in enumerate(entries):
            newelem = elem.replace('\n', '').strip()
            if entlen > 1:
                if i == 0:
                    newelem = newelem + ']'
                elif i == entlen-1:
                    newelem = '[' + newelem
                else:
                    newelem = '[' + newelem + ']'

            jelem = None
            try:
                jelem = json.loads(newelem)
            except json.decoder.JSONDecodeError:
                # This is a bug in the output of FRR's show evpn vni when
                # there's no EVPN
                if 'show evpn vni detail json' in newelem:
                    newelem = newelem[:-1]
                    jelem = json.loads(newelem)

            if jelem is None:
                self.logger.error(f"Unable to decode JSON in file {file}")
                continue

            if not all(key in required_keys for key in jelem[0].keys()):
                self.logger.error(
                    f'Ignoring entry with missing required key fields {jelem}')
                continue

            key = jelem[0]['cmd'].split('|')[0].strip()

            if key not in self.data:
                self.data[key] = {'curpos': 0, 'entry': []}

            self.data[key]['entry'].append(jelem)

    def post_commands(self, service_callback, svc_defn: dict,
                      cb_token: RsltToken):
        '''Post command outputs back to service that requested them'''

        if cb_token:
            cb_token.nodeQsize = self._service_queue.qsize()
        self._service_queue.put_nowait([service_callback, svc_defn, cb_token])

    async def run(self):
        '''Main workhorse routine, serving data from files based on command'''

        while True:
            request = await self._service_queue.get()
            # request consists of callback fn, service_defn string, cb_token
            if request:
                if not await self.exec_service(request[0], request[1],
                                               request[2]):
                    return

    def _create_result(self, cmd, status, data) -> dict:
        '''Create result object to be posted back to service object'''

        result = {
            "status": status,
            "timestamp": int(datetime.now(tz=timezone.utc).timestamp() * 1000),
            "cmd": cmd,
            "devtype": 'Unknown',
            "namespace": 'Unknown',
            "hostname": self.hostname,
            "address": '0.0.0.0',
            "version": 0,
            "data": data,
        }
        return result

    async def exec_service(self, service_callback, svc_defn: dict,
                           cb_token: RsltToken) -> bool:
        '''This looks up the command, returns the output if available'''

        def _add_cmd_to_cmdset(cmd: str, cmdset: set) -> None:
            """Add a command to the set of commands to be executed"""
            if isinstance(cmd, list):
                # the first command is what we need
                subcmd = cmd[0]['command'].split('|')[0].strip()
                cmdset.add(subcmd)
            else:
                cmdset.add(cmd.split('|')[0].strip())

        cmdset = set()
        for device in svc_defn.keys():
            defn = svc_defn[device]
            if isinstance(defn, list):
                for elem in defn:
                    cmd = elem.get('command', None)
                    if not cmd:
                        continue
                    _add_cmd_to_cmdset(cmd, cmdset)
            else:
                cmd = defn.get('command', None)
                if not cmd:
                    continue

                _add_cmd_to_cmdset(cmd, cmdset)

        nodata = True
        for cmd in cmdset:
            data_entry = self.data.get(cmd, {'entry': []})
            for _ in range(len(data_entry['entry'])):
                retdata = data_entry['entry'].pop()
                await service_callback(retdata, cb_token)
                nodata = False
        if nodata:
            await service_callback([self._create_result(
                cmd, HTTPStatus.NO_CONTENT, [])], cb_token)

        return True



from itertools import zip_longest

import pandas as pd
import streamlit as st

from suzieq.gui.stlit.pagecls import SqGuiPage


class PathDebugPage(SqGuiPage):
    '''Page for Path trace page'''
    _title = 'Path-Debug'
    _state = None

    @property
    def add_to_menu(self) -> bool:
        return False

    def build(self):
        self._create_sidebar()
        self._render(None)

    # pylint: disable=protected-access
    def set_path_state(self, state):
        '''Save the path state from the Path window'''
        try:
            self._state = state._old_state['pages']['Path']
        except KeyError:
            self._state = state._new_session_state['pages']['Path']

    def _get_state_from_url(self):
        pass

    def _create_sidebar(self) -> None:
        st.sidebar.markdown("""Displays information from the various tables """
                            """used to build path""")

    def _create_layout(self) -> None:
        pass

    def _render(self, _) -> None:
        url_params = st.experimental_get_query_params()
        if url_params.get('lookupType', 'hop') == ['hop']:
            self._handle_hop_url(url_params)
        else:
            self._handle_edge_url(url_params)

    def _sync_state(self) -> None:
        pass

    def _save_page_url(self) -> None:
        pass

    # pylint: disable=too-many-statements
    def _handle_edge_url(self, url_params: dict):
        '''Display tables associated with a link'''

        namespace = url_params.get('namespace', [""])[0]
        hostname = url_params.get('hostname', [""])[0]
        nhip = url_params.get('nhip', [""])[0]
        ipLookup = url_params.get('ipLookup', [""])[0]
        vtepLookup = url_params.get('vtepLookup', [""])[0]
        vrf = url_params.get('vrf', [""])[0]
        ifhost = url_params.get('ifhost', [""])[0]
        macaddr = url_params.get('macaddr', [""])[0]
        oif = url_params.get('oif', [""])[0]

        state = self._state._state
        if not hostname:
            st.error('No hostname found to display information for')
            st.stop()

        st.header(f'Debug Tables for Path from {state.source} to '
                  f'{state.dest}')
        hoptype = 'Bridged' if macaddr else 'Routed'
        st.subheader(f'{hoptype} hop between {hostname} and {ifhost}')

        pathobj = getattr(self._state, '_pathobj', None)
        engobj = pathobj.engine

        if ipLookup:
            if not vtepLookup or (ipLookup != vtepLookup):
                st.info(f'Route Lookup on {hostname}')
                st.dataframe(data=engobj._rdf.query(
                    f'hostname=="{hostname}" and vrf=="{vrf}"'))

            if vtepLookup:
                st.info(f'Underlay Lookup on {hostname} for {vtepLookup}')
                vtepdf = engobj._underlay_dfs.get(vtepLookup,
                                                  pd.DataFrame())
                if not vtepdf.empty:
                    st.dataframe(data=vtepdf.query(
                        f'hostname=="{hostname}" and vrf=="default"'))
            if nhip:
                st.info(
                    f'ARP/ND Table on {hostname} for nexthop {nhip}, '
                    f'oif={oif}')
                arpdf = engobj._arpnd_df.query(f'hostname=="{hostname}" and '
                                               f'ipAddress=="{nhip}" and '
                                               f'oif=="{oif}"')
                st.dataframe(data=arpdf)

                if ':' in nhip:
                    dropcol = ['ipAddressList']
                else:
                    dropcol = ['ip6AddressList']
                if not arpdf.empty:
                    nhmac = arpdf.macaddr.iloc[0]
                    if nhmac:
                        if_df = engobj._if_df.query(f'macaddr=="{nhmac}" and '
                                                    f'hostname=="{ifhost}"') \
                                             .drop(columns=dropcol)
                        label = (f'matching nexthop {nhip}, '
                                 f'macaddr {nhmac} on '
                                 f'host {ifhost}')
                    else:
                        label = f'matching nexthop {nhip} on host {ifhost}'
                        if_df = engobj._if_df.query(f'hostname=="{ifhost}"') \
                                             .drop(columns=dropcol)
                else:
                    if_df = engobj._if_df.query(f'hostname=="{ifhost}"')\
                        .drop(columns=dropcol)
                    label = f'matching host {ifhost}'
                if nhip != '169.254.0.1':
                    st.info(f'Interfaces {label}')
                    s = if_df.ipAddressList.str \
                                           .startswith(f'{nhip}/') \
                                           .dropna()
                    s = s.loc[s]
                    st.dataframe(
                        data=engobj._if_df.iloc[s.loc[s].index])
                else:
                    st.info(f'Interfaces {label}')
                    st.dataframe(data=if_df)
        if macaddr:
            with st.expander(f'MAC Table for {hostname}, MAC addr {macaddr}',
                             expanded=True):
                st.dataframe(data=pathobj.engine._macsobj.get(
                    namespace=namespace, hostname=hostname, macaddr=macaddr))

    # pylint: disable=protected-access, too-many-statements

    def _handle_hop_url(self, url_params):
        '''Handle table display associated with hop'''

        namespace = url_params.get('namespace', [""])[0]
        hostname = url_params.get('hostname', [""])[0]

        state = self._state._state
        if not hostname:
            st.error('No hostname found to display information for')
            st.stop()

        st.header(f'Debug Tables for Path from {state.source} to '
                  f'{state.dest}')

        failed_dfs = getattr(self._state, '_failed_dfs', None)
        for tbl, fdf in failed_dfs.items():
            if not fdf.empty:
                this_df = fdf.query(f'namespace=="{namespace}" and '
                                    f'hostname=="{hostname}"')
                if this_df.empty:
                    continue
                table_expander = st.expander(
                    f'Failed {tbl} Table', expanded=not fdf.empty)
                table_expander.dataframe(fdf)

        pathobj = getattr(self._state, '_pathobj', None)
        df = getattr(self._state, '_path_df', None)
        engobj = pathobj.engine

        if df.empty:
            st.warning('Empty path dataframe')
            st.stop()

        host_dfg = df.query(f'hostname == "{hostname}"') \
                     .groupby(by=['hopCount'])

        df2 = host_dfg.agg({'vrf': ['unique'], 'ipLookup': ['unique'],
                            'nexthopIp': ['unique'], 'oif': ['unique'],
                            'macLookup': ['unique'],
                            'vtepLookup': ['unique']}).reset_index()
        df2.columns = ['hopCount', 'vrf', 'ipLookup', 'nexthopIp', 'oif',
                       'macaddr', 'vtepLookup']
        df2 = df2.explode('hopCount').explode('vrf').explode('ipLookup') \
                                                    .explode('macaddr') \
                                                    .explode('vtepLookup')
        df2.drop_duplicates(subset=['vrf', 'ipLookup'], inplace=True)

        for row in df2.itertuples():
            with st.expander(f'Lookups on {hostname}, for hopcount: '
                             f'{row.hopCount}', expanded=True):
                if row.macaddr:
                    st.info(f'MAC Table on {hostname}, MAC addr {row.macaddr}')
                    # pylint: disable=protected-access
                    st.dataframe(data=engobj._macsobj.get(namespace=namespace,
                                                          hostname=hostname,
                                                          macaddr=row.macaddr))
                    continue

                if (row.ipLookup != row.vtepLookup):
                    st.info(f'Route Lookup on {hostname}')
                    # pylint: disable=protected-access
                    st.dataframe(data=engobj._rdf.query(
                        f'hostname=="{hostname}" and vrf=="{row.vrf}"'))

                if row.vtepLookup:
                    st.info(
                        f'Underlay Lookup on {hostname} for {row.vtepLookup}')
                    # pylint: disable=protected-access
                    vtepdf = engobj._underlay_dfs.get(row.vtepLookup,
                                                      pd.DataFrame())
                    if not vtepdf.empty:
                        st.dataframe(data=vtepdf.query(
                            f'hostname=="{hostname}" and vrf=="default"'))

                oifs = row.oif.tolist()
                nhops = row.nexthopIp.tolist()
                prev_nhop = ''
                for oif, nhop in zip_longest(oifs, nhops):
                    _, arpcol = st.columns([1, 40])
                    _, ifcol = st.columns([2, 40])
                    # this logic because I don't know what fn to use with agg
                    # above to not remove non-unique nhop.
                    if not nhop and prev_nhop:
                        nhop = prev_nhop
                    else:
                        prev_nhop = nhop
                    arpdf = engobj._arpnd_df.query(
                        f'hostname=="{hostname}" '
                        f' and ipAddress=="{nhop}" and '
                        f'oif=="{oif}"')
                    with arpcol:
                        st.info(f'ARP/ND Lookup on {hostname} for {nhop}')
                        st.dataframe(data=arpdf, height=100)

                    if not arpdf.empty:
                        if ':' in nhop:
                            dropcol = ['ipAddressList']
                        else:
                            dropcol = ['ip6AddressList']
                        if nhop == '169.254.0.1':
                            macaddr = arpdf.macaddr.iloc[0]
                            if_df = engobj._if_df.query(
                                f'macaddr=="{macaddr}"') \
                                .drop(columns=dropcol)
                            label = f'matching nexthop {nhop}, '\
                                f'macaddr {macaddr}'
                        else:
                            if_df = engobj._if_df.drop(columns=dropcol)
                            label = f'matching nexthop {nhop}'
                    else:
                        label = f'matching nexthop {nhop}'
                        if_df = engobj._if_df
                    if ':' in nhop:
                        s = if_df.ip6AddressList \
                                 .explode() \
                                 .str.startswith(f'{nhop}/').dropna()
                        s = s.loc[s]
                        if_df = if_df.iloc[s.loc[s].index]
                    elif nhop != '169.254.0.1':
                        s = if_df.ipAddressList \
                                 .explode() \
                                 .str.startswith(f'{nhop}/').dropna()
                        s = s.loc[s]
                        if_df = if_df.iloc[s.loc[s].index]
                    with ifcol:
                        st.info(f'Interfaces {label}')
                        st.dataframe(data=if_df, height=600)
            st.markdown("<hr>", unsafe_allow_html=True)



from suzieq.engines.pandas.engineobj import SqPandasEngine


class TopcpuObj(SqPandasEngine):
    '''Backend class to handle manipulating topcpu table with pandas'''

    @staticmethod
    def table_name():
        '''Table name'''
        return 'topcpu'



"""Module containing base class for inventorySource plugins
"""
import asyncio
from copy import copy, deepcopy
from os.path import isfile
from pathlib import Path
from typing import Dict, List


from suzieq.poller.controller.base_controller_plugin import \
    ControllerPlugin, InventoryPluginModel
from suzieq.poller.controller.credential_loader.base_credential_loader import \
    CredentialLoader, check_credentials
from suzieq.poller.controller.utils.inventory_utils import read_inventory
from suzieq.shared.exceptions import InventorySourceError


class SourceModel(InventoryPluginModel):
    """Model for inventory source validation

    IMPORTANT:
    do not use namespace, device and auth as field names
    """


class Source(ControllerPlugin):
    """Base class for plugins which reads inventories"""

    def __init__(self, input_data, validate: bool = True) -> None:
        super().__init__(input_data, validate)

        self._inventory = {}
        self._inv_is_set = False
        self._inv_is_set_event = asyncio.Event()
        self._auth = input_data.pop('auth', None)  # auth object
        self._device = input_data.pop('device', {})
        self._namespace = input_data.pop('namespace')
        self._run_once = input_data.pop('run_once', False)
        self._data = None

        self._inv_format = [
            'address',
            'namespace',
            'port',
            'transport',
            'devtype',
            'hostname',
            'jump_host',
            'jump_host_key_file',
            'ignore_known_hosts',
            'slow_host',
        ]

        self._load(input_data)

    @property
    def name(self) -> str:
        """Name of the source set in the inventory file

        Returns:
            str: name of the source
        """
        return self._data.name

    @classmethod
    def default_type(cls) -> str:
        return 'native'

    @classmethod
    def get_data_model(cls):
        return SourceModel

    def _load(self, input_data):
        """Store informations from raw data"""
        if self._validate:
            self._data = self.get_data_model()(**input_data)
        else:
            self._data = self.get_data_model().construct(**input_data)
        if not self._data:
            raise InventorySourceError(
                'input_data was not loaded correctly')

    async def get_inventory(self) -> Dict:
        """Retrieve the inventory from the source. If the inventory is not
        ready the function will wait until it is.

        Returns:
            List[Dict]: the device inventory from the source
        """
        # If the inventory is not ready wait until it is
        if not self._inv_is_set:
            await self._inv_is_set_event.wait()

        inventory_snapshot = copy(self._inventory)

        return inventory_snapshot

    def set_inventory(self, new_inventory: Dict):
        """Set the inventory in a thread safe way

        The function will try to set the inventory until the timeout
        expires.

        Before setting the new inventory, it calls the validator.

        Args:
            new_inventory ([List[Dict]]): the new inventory to set
        """
        if self._auth:
            self._auth.load(new_inventory)
        else:
            check_credentials(new_inventory)
        self.set_device(new_inventory)
        missing_keys = self._is_invalid_inventory(new_inventory)
        if missing_keys:
            raise InventorySourceError(
                f'{self.name} missing informations: {missing_keys}')

        self._inventory = new_inventory

        # If the inventory has been set for the first time, we need to
        # unlock who is waiting for it
        if not self._inv_is_set:
            # the inventory has been set for the first time
            self._inv_is_set = True
            self._inv_is_set_event.set()

    def _is_invalid_inventory(self, inventory: Dict) -> List[str]:
        """Validate the inventory

        The goal of this function is to check that all the devices has all
        the values in self._inv_format

        This function is called inside the set_inventory

        Args:
            inventory (Dict): inventory to validate

        Returns:
            List[str]: list of missing fields
        """
        for host in inventory.values():
            host_keys = set(self._inv_format)
            for key in host.keys():
                if key in host_keys:
                    host_keys.remove(key)

            if host.get('transport') == 'https' and not host.get('devtype'):
                raise InventorySourceError('Missing devtype in https transport'
                                           f' for host {host.get("address")}')

            if host_keys:
                return list(host_keys)

        return []

    @classmethod
    def init_plugins(cls, plugin_conf: Dict, validate: bool = False)\
            -> List[Dict]:
        """This method is overrided because sources is different from other
        plugins.
        From the fields 'path', this function is going to load more than
        one source.
        The generate function for all other plugin will simply read the
        configuration and instance an object

        Args:
            plugin_conf (Dict): source plugin configuration dictionary.
            Must contain 'path' key with the inventory file as value.

        Raises:
            InventorySourceError: No 'path' key in plugin_conf
            RuntimeError: Unknown plugin

        Returns:
            List[Dict]: [description]
        """
        src_plugins = []
        plugin_classes = cls.get_plugins()
        if not plugin_conf.get('path'):
            raise InventorySourceError('A source plugin cannot be initialized'
                                       'without the inventory file path')
        src_confs = _load_inventory(plugin_conf.get('path'))
        run_once = plugin_conf.get('single-run-mode', None)
        for src_conf in src_confs:
            ptype = src_conf.get('type') or Source.default_type()

            if ptype not in plugin_classes:
                raise RuntimeError(
                    f'Unknown plugin called {ptype}'
                )
            src_conf.update({'run_once': run_once})
            src_plugins.append(plugin_classes[ptype](src_conf, validate))
        return src_plugins

    def set_device(self, inventory: Dict[str, Dict]):
        """Add device config from inventory file to the inventory

        Args:
            inventory (Dict[str,Dict]): inventory
        """
        jump_host = None
        jump_host_key_file = None
        transport = None
        ignore_known_hosts = False
        slow_host = False
        port = None
        devtype = None

        if self._device:
            jump_host = self._device.get('jump-host')
            if jump_host and not jump_host.startswith("//"):
                jump_host = f'//{jump_host}'
            jump_host_key_file = self._device.get('jump-host-key-file')
            if jump_host_key_file and \
                    not Path(jump_host_key_file).is_file():
                raise InventorySourceError(f'{self.name} Jump host key file'
                                           f" at {jump_host_key_file} doesn't"
                                           " exists")
            transport = self._device.get('transport')
            if transport:
                # get the string from the enum
                transport = transport.value
            ignore_known_hosts = self._device.get('ignore-known-hosts', False)
            slow_host = self._device.get('slow-host', False)
            port = self._device.get('port')
            devtype = self._device.get('devtype')

        for node in inventory.values():
            node.update({
                'jump_host': node.get('jump_host') or jump_host,
                'jump_host_key_file': node.get('jump_host_key_file')
                or jump_host_key_file,
                'ignore_known_hosts': node.get('ignore_known_hosts')
                or ignore_known_hosts,
                'transport': node.get('transport') or transport or 'ssh',
                'port': node.get('port') or port or 22,
                'devtype': node.get('devtype') or devtype,
                'slow_host': node.get('slow_host', '') or slow_host,
            })

    def _validate_device(self):
        if self._device:
            dev_fields = ['name', 'jump-host', 'jump-host-key-file',
                          'ignore-known-hosts', 'transport', 'port',
                          'slow-host', 'devtype']
            inv_fields = [x for x in self._device if x not in dev_fields]
            if inv_fields:
                raise InventorySourceError(
                    f'{self._device.get("name")}: Unknown fields called '
                    f'{inv_fields}')


def _load_inventory(source_file: str) -> List[dict]:
    """Load inventory from a file

    Args:
        source_file (str): inventory file

    Raises:
        InventorySourceError: inventory file doesn't exists
        or invalid

    Returns:
        List[dict]: list of sources
    """
    if not isfile(source_file):
        raise InventorySourceError(f"File {source_file} doesn't exists")

    inventory = read_inventory(source_file)

    ns_list = inventory.get('namespaces')
    sources_dict = inventory.get('sources')
    auths_dict = inventory.get('auths', {})
    devs_dict = inventory.get('devices', {})

    sources = []
    for ns in ns_list:
        source_name = ns.get('source')
        namespace = ns.get('name')

        source = sources_dict.get(source_name)

        if ns.get('auth'):
            # this is only a temporary fix, in future releases I will move the
            # credential loader initialization outside of this function.
            auth = deepcopy(auths_dict.get(ns['auth']))
            auth_type = auth.get('type') or CredentialLoader.default_type()
            if "-" in auth_type:
                auth_type = auth_type.replace("-", "_")
            auth['type'] = auth_type
            source['auth'] = CredentialLoader.init_plugins(auth)[0]
        else:
            source['auth'] = None

        if ns.get('device'):
            device = devs_dict.get(ns['device'])
            source['device'] = device
        else:
            source['device'] = None

        source['namespace'] = namespace

        sources.append(source)

    return sources



"""
This module contains all the common logic of the
Suzieq poller inventory sources
"""
import abc
import asyncio
import logging
from collections import defaultdict
from typing import Callable, Coroutine, Dict, List

from suzieq.poller.worker.nodes.node import Node
from suzieq.shared.exceptions import SqPollerConfError
from suzieq.shared.sq_plugin import SqPlugin

logger = logging.getLogger(__name__)


class Inventory(SqPlugin):
    """Inventory is the base class implemented by the
    inventory sources, providing the list of devices
    to poll.
    """

    def __init__(self, add_task_fn: Callable, **kwargs) -> None:
        """Instantiate the Inventory class

        Args:
            add_task_fn (Callable): the function to call to schedule
                a task in the poller.
        """
        self._nodes = {}
        self._node_tasks = {}
        self.add_task_fn = add_task_fn

        self.connect_timeout = kwargs.pop('connect_timeout', 15)
        self.ssh_config_file = kwargs.pop('ssh_config_file', None)

    @property
    def nodes(self) -> Dict[str, Node]:
        """Get the current list of nodes in the inventory

        Returns:
            Dict[str, Node]: get a dictionary with the nodes in the inventory
                with format {namespace.hostname: Node}
        """
        return self._nodes

    @property
    def running_nodes(self) -> List[Coroutine]:
        """Get the the couroutines of the running nodes

        Returns:
            List[Coroutine]: a list containing the coroutines of the running
                nodes
        """
        return self._node_tasks

    async def build_inventory(self) -> Dict[str, Node]:
        """Retrieve the list of nodes to poll and instantiate
        all the Nodes objects in the retrieved inventory.

        Raises:
            SqPollerConfError: in case of wrong inventory configuration
            InventorySourceError: in case of error with the inventory source

        Returns:
            Dict[str, Node]: a list containing all the nodes in the inventory
        """

        inventory_list = await self._get_device_list()
        if not inventory_list:
            raise SqPollerConfError('The inventory source returned no hosts')

        # Initialize the nodes in the inventory
        self._nodes = await self._init_nodes(inventory_list)
        return self._nodes

    def get_node_callq(self) -> Dict[str, Dict]:
        """Get the dictionary allowing to send command query
        on the nodes

        Returns:
            Dict[str, Dict]: a dictionary having, for each 'namespace.hostname'
                the node hostname and the function allowing to call
                a query on the node.
        """
        node_callq = defaultdict(lambda: defaultdict(dict))

        node_callq.update({x: {'hostname': self._nodes[x].hostname,
                               'postq':    self._nodes[x].post_commands}
                           for x in self._nodes})

        return node_callq

    async def schedule_nodes_run(self):
        """Schedule the nodes tasks in the poller, so that they can
        start processing the command query passed to the nodes.

        This function should be called only once, if called more
        than once, it won't have any effect.
        """
        if not self._node_tasks:
            self._node_tasks = {node: self._nodes[node].run()
                                for node in self._nodes}
            await self.add_task_fn(list(self._node_tasks.values()))

    async def _init_nodes(self, inventory_list:
                          List[Dict]) -> Dict[str, Node]:
        """Initialize the Node objects given the of credentials of the nodes.
        After this function is called, a connection with the nodes in the list
        is performed.

        Returns:
            Dict[str, Node]: a list containing the initialized Node objects
        """
        init_tasks = []
        nodes_list = {}

        for host in inventory_list:
            new_node = Node()
            init_tasks += [new_node.initialize(
                **host,
                connect_timeout=self.connect_timeout,
                ssh_config_file=self.ssh_config_file
            )]

        for n in asyncio.as_completed(init_tasks):
            try:
                newnode = await n
            except Exception as e:  # pylint: disable=broad-except
                logger.error(
                    f'Encountered error {e} in initializing node')
                continue
            if newnode.devtype == "unsupported":
                logger.error(
                    f'Unsupported device type for '
                    f'{newnode.address}.{newnode.port}'
                )
            elif not newnode.devtype:
                logger.warning('Unable to determine device type for '
                               f'{newnode.address}. Will retry')
            else:
                logger.info(f'Added node {newnode.hostname}:{newnode.port} '
                            f'of type {newnode.devtype}')
            nodes_list.update({self.get_node_key(newnode): newnode})

        return nodes_list

    @ abc.abstractmethod
    async def _get_device_list(self) -> List[Dict]:
        """Retrieve the devices credentials from the inventory
        source

        Raises:
            InventorySourceError: in case of error with the inventory source

        Returns:
            List[Dict]: the list of the credentials of the devices
                in the Suzieq native inventory file.
        """
        raise NotImplementedError

    @ staticmethod
    def get_node_key(node: Node) -> str:
        """Given a node object it returns its ID key

        Args:
            node (Node): the node from which retrieving the key

        Returns:
            str: a string containing the ID key of the node
        """
        return f'{node.nsname}.{node.hostname}'



from ipaddress import ip_address
import re
import pandas as pd

from suzieq.sqobjects.basicobj import SqObject
from suzieq.shared.utils import (humanize_timestamp,
                                 convert_macaddr_format_to_colon)


class NetworkObj(SqObject):
    '''The object providing access to the virtual table: network'''

    def __init__(self, **kwargs):
        super().__init__(table='network', **kwargs)
        self._valid_get_args = ['namespace', 'hostname', 'version', 'os',
                                'model', 'vendor', 'columns', 'query_str']
        self._valid_find_args = ['namespace', 'hostname', 'vrf', 'vlan',
                                 'address', 'query_str']
        self._unique_def_column = ['namespace']

    def find(self, **kwargs) -> pd.DataFrame():
        '''Find network attach point for a given address'''

        addresses = kwargs.get('address', '')

        if not self.ctxt.engine:
            raise AttributeError('No analysis engine specified')

        if not addresses:
            raise AttributeError('Must specify address')

        for addr in addresses:
            try:
                ip_address(addr)
            except ValueError:
                addr = convert_macaddr_format_to_colon(addr)
                if not re.match(
                        "[0-9a-f]{2}([-:]?)[0-9a-f]{2}(\\1[0-9a-f]{2}){4}$",
                        addr):
                    return pd.DataFrame(
                        {'error': [f'Not valid IP or MAC address: {addr}']})
        try:
            self._check_input_for_valid_args(self._valid_find_args, **kwargs)
        except (ValueError, AttributeError) as error:
            df = pd.DataFrame({'error': [f'{error}']})
            return df

        return self.engine.find(**kwargs)

    def humanize_fields(self, df: pd.DataFrame, _=None) -> pd.DataFrame:
        '''Humanize the timestamp fields'''
        if df.empty:
            return df

        if 'lastUpdate' in df.columns:
            df['lastUpdate'] = humanize_timestamp(df.lastUpdate,
                                                  self.cfg.get('analyzer', {})
                                                  .get('timezone', None))

        return super().humanize_fields(df)



"""
This module contains the logic needed to parse the Suzieq native inventory
file
"""

import re
import sys
from dataclasses import dataclass
from ipaddress import ip_address
from pathlib import Path
from typing import Dict, List

import yaml
from pydantic import ValidationError
from suzieq.poller.controller.base_controller_plugin import ControllerPlugin
from suzieq.poller.controller.utils.inventory_models import (DeviceModel,
                                                             InventoryModel,
                                                             NamespaceModel)
from suzieq.shared.exceptions import InventorySourceError


@dataclass
class InventoryValidationError:
    """Dataclass for inventory errors handling.

    location: namespaces, sources, devices, auths
    name: name of the object raising an exception
    what: the raised exception
    """
    location: str
    name: str
    what: Exception


def read_inventory(source_file: str) -> Dict:
    """Read the Suzieq native inventory returning a dictionary with its
    content.

    Args:
        source_path (str): the source file

    Raises:
        InventorySourceError: raised if there is any error if the file parsing

    Returns:
        Dict: a dictionary containing the content of the native inventory file
    """
    if not Path(source_file).is_file():
        raise InventorySourceError(
            f"Inventory file {source_file} doesn't exits")

    inventory_data = {}
    with open(source_file, 'r') as fp:
        file_content = fp.read()
        try:
            inventory_data = yaml.safe_load(file_content)
        except Exception as e:
            raise InventorySourceError(f'Invalid Suzieq inventory file: {e}')

    return validate_raw_inventory(inventory_data)

    # for k in inventory:
    #     inventory[k] = get_inventory_config(k, inventory_data)

    # inventory['namespaces'] = inventory_data.get('namespaces')

    # return inventory


def validate_raw_inventory(inventory: dict) -> Dict:
    """Validate the inventory read from file and update its content
    to be easier to load into plugins

    Args:
        inventory (dict): inventory read from file

    Returns:
        Dict: updated inventory

    Raises:
        InventorySourceError: invalid inventory
    """
    # pylint: disable=too-many-nested-blocks
    # pylint: disable=broad-except

    def find_error(
            errors: List[InventoryValidationError],
            name: str, location: str) -> bool:
        """Checks if the name is already in errors
        It is used to avoid launching chained exceptions

        Args:
            errors (List[InventoryValidationError]): list of errors found
            name (str): name of the object who generated the error
            location (str): location of the name

        Returns:
            bool: True if the name is found
        """
        # pylint: disable=used-before-assignment
        # I think it is a bug of pylint. It tells me that e.name
        # is used before assignement
        return any((True for e in errors
                    if (e.location == location and e.name == name)))

    if not inventory:
        raise InventorySourceError('The inventory is empty')

    if not isinstance(inventory, dict):
        raise InventorySourceError(
            'Expected mapping instead of list. Check the docs to know how to '
            'build the inventory file')

    try:
        inv_model = InventoryModel(**inventory)
    except ValidationError as e:
        output_inv_errors([
            InventoryValidationError(
                location='inventory',
                name='',
                what=e
            )])
    new_inventory = {k: {} for k in inv_model.dict()}
    base_plugins = ControllerPlugin.get_plugins()
    errors: List[InventoryValidationError] = []
    for mf, fields in inv_model.dict().items():
        if not fields:
            # 'devices' and 'auths' can be omitted if not needed
            continue
        if mf == 'namespaces':
            # namespaces are validated later
            continue

        global_specs = {}
        for plugin_specs in fields:
            try:
                if 'copy' in plugin_specs:
                    # copy the content of the other inventory
                    # into the current inventory and override
                    # values
                    copyname = plugin_specs['copy']
                    if copyname not in global_specs:
                        if not find_error(errors, copyname, mf):
                            raise InventorySourceError(
                                "copy value must be a 'name' of an already "
                                f"defined {mf}: {copyname} not found"
                            )
                        else:
                            # an error related to 'copyname' was already raised
                            continue
                    plugin_specs = copy_inventory_item(
                        global_specs[copyname]['specs'], plugin_specs)

                if mf == 'auths':
                    # I don't like this step, but there is no other way
                    # to validate credentials otherwise
                    # In a future release I will rename credential_loader to
                    # auth and get rid of this if
                    mtype = 'credential_loader'
                else:
                    # remove the trailing 's'
                    mtype = mf[:-1]

                if mtype in base_plugins:
                    # validate 'auths' and 'sources'
                    base_plugin_class = base_plugins[mtype]
                    ptype = plugin_specs.get(
                        'type', base_plugin_class.default_type())\
                        .replace('-', '_')
                    plugin_class = base_plugin_class.get_plugins(ptype)\
                        .get(ptype)
                    if not plugin_class:
                        raise InventorySourceError(
                            f'Unknown {mf} of type {ptype}')
                    try:
                        model_class = plugin_class.get_data_model()
                    except NotImplementedError:
                        raise InventorySourceError(
                            f'Data model not implemented for {ptype}')
                    validated_obj = model_class(**plugin_specs)
                    name = validated_obj.name

                elif mf == 'devices':
                    # validate 'devices'
                    validated_obj = DeviceModel(**plugin_specs)
                    name = validated_obj.name

                if name in global_specs:
                    raise InventorySourceError(
                        f'{mf}.{name} is not unique')
                global_specs[name] = {
                    'validated': validated_obj.dict(by_alias=True),
                    'specs': plugin_specs
                }
            except Exception as e:
                errors.append(
                    InventoryValidationError(
                        location=mf,
                        name=plugin_specs.get('name'),
                        what=e
                    )
                )
        new_inventory[mf] = {name: g['validated']
                             for name, g in global_specs.items()}

    # validate 'namespaces'
    global_ns_specs = []
    for ns_specs in inv_model.namespaces:
        try:
            ns = NamespaceModel(**ns_specs)
        except Exception as e:
            errors.append(
                InventoryValidationError(
                    location='namespaces',
                    name=ns_specs.get('name'),
                    what=e
                )
            )
            continue
        for field, value in ns.dict().items():
            if field == 'name':
                continue
            inv_field = field + 's'
            try:
                if value and value not in new_inventory[inv_field]:
                    if find_error(errors, value, inv_field):
                        # an error already raised for this object
                        continue
                    raise InventorySourceError(
                        f"No {field} called '{value}'")
            except Exception as e:
                errors.append(
                    InventoryValidationError(
                        location='namespaces',
                        name=ns_specs.get('name'),
                        what=e
                    )
                )

        global_ns_specs.append(ns.dict(by_alias=True))
    new_inventory['namespaces'] = global_ns_specs

    if errors:
        output_inv_errors(errors)

    return new_inventory


def output_inv_errors(errors: List[InventoryValidationError]):
    """Creates a good looking output for inventory errors

    Args:
        errors (List[Exception]): list of errors found
    """

    error_str = '\n'
    for e in errors:
        error_str += f'- {e.location}'
        if e.name:
            error_str += f'.{e.name}'
        error_str += ':\n'
        if isinstance(e.what, ValidationError):
            for ve in e.what.errors():
                field = '.'.join(map(str, ve.get('loc', [''])))
                error_str += f"  + {field}: {ve.get('msg','')}\n"
        else:
            error_str += f'  + {str(e.what)}\n'
        error_str += '\n'
    raise InventorySourceError(error_str)


def copy_inventory_item(orig: Dict, dest: Dict) -> Dict:
    """Copy the content of orig inside dest for parameters not
    specified in dest

    Args:
        orig (Dict): original dict to copy
        dest (Dict): destination dict to update

    Returns:
        Dict: updated destination dict
    """
    for k, v in orig.items():
        if k not in dest:
            dest[k] = v
    if 'copy' in dest:
        dest.pop('copy')
    return dest


def validate_hostname(in_hostname: str) -> bool:
    """Validates if the hostname is an ip address or a domain

    Args:
        in_hostname (str): hostname to evaluate

    Returns:
        bool: validation result
    """
    try:
        ip_address(in_hostname)
        if sys.version_info[:3] < (3, 8, 12):
            # for python versions before 3.8.12, the ip_address function
            # doesn't tolerate addresses having octets with leading 0, like
            # "10.00.0.1"
            # The following regex makes the validation consistent across
            # python versions
            wrong_ip_pattern = r'(\.0\d+\.?)|(^0\d+\.?)'
            if re.search(wrong_ip_pattern, in_hostname):
                return False
        return True
    except ValueError:
        # check if it was a misspelled ipv4
        ipv4_pattern = r'^(\d+\.?)+$'
        if re.fullmatch(ipv4_pattern, in_hostname):
            # if this pattern matches, it means that the function
            # ip_address raise an error due to the fact of an misspelled ipv4
            # e.g. 10.0.1 or 10.0.2555.1
            return False

        # validate the domain name
        pattern = r'^([a-zA-Z0-9-.]{1,253})?$'
        m = re.fullmatch(pattern, in_hostname)

        if not m:
            return False

        hostname = m.groups()[0]

        # strip trailing dot if any
        hostname = hostname.strip('.')

        if len(hostname) > 253:
            return False

        for label in hostname.split('.'):
            lm = re.fullmatch(r'^([a-zA-Z0-9-]+)$', label)
            if lm and len(label) < 64 and \
                    not (label.startswith('-') or label.endswith('-')):
                continue
            else:
                return False
        return True



import os
import sys
from collections import defaultdict
import base64
from typing import Dict

import streamlit as st

from suzieq.gui.stlit.guiutils import (SUZIEQ_COLOR, get_image_dir,
                                       get_main_session_by_id, SuzieqMainPages)
from suzieq.gui.stlit.pagecls import SqGuiPage
from suzieq.version import SUZIEQ_VERSION
from suzieq.shared.utils import sq_get_config_file


def set_horizontal_radio():
    '''Make the radio buttons horizontal'''
    st.write('<style>div.row-widget.stRadio > '
             'div{flex-direction:row;}</style>',
             unsafe_allow_html=True)


def set_vertical_radio():
    '''Make the radio buttons horizontal'''
    st.write('<style>div.row-widget.stRadio > '
             'div{flex-direction:column;}</style>',
             unsafe_allow_html=True)


def display_help_icon(url: str):
    '''Display Help Icon with click to take you to appropriate page'''

    help_img = f'{get_image_dir()}/helps.png'
    with open(help_img, "rb") as f:
        img = base64.b64encode(f.read()).decode()

    st.sidebar.markdown(
        f'<a target="_help" href="{url}"><img class="help-img" '
        f'src="data:image/png;base64,'
        f'{img}"></a>',
        unsafe_allow_html=True)


def display_title(page: str):
    '''Render the logo and the app name'''

    state = st.session_state
    menulist = state.menulist
    search_text = state.search_text

    LOGO_IMAGE = f'{get_image_dir()}/logo-small.jpg'
    st.markdown(
        """
        <style>
            .logo-container {
                display: flex;
            }
            .logo-text {
                color: %s;
            }
            .logo-img {
                max-height: 7.5em;
                margin-bottom: 1em;
                float:right;
            }
            @media(max-width: 960px){
                .logo-container {
                    justify-content: center;
                }
                .logo-text {
                    display: none;
                }
            }
            @media(max-width: 640px){
                .logo-text {
                    display: inline;
                }
            }
        </style>
        """ % (SUZIEQ_COLOR),
        unsafe_allow_html=True
    )

    title_col, _, page_col, srch_col = st.columns([2, 1, 2, 2])

    with open(LOGO_IMAGE, "rb") as f:
        img = base64.b64encode(f.read()).decode()

    with title_col:
        st.markdown(
            f"""
            <div class="logo-container">
                <img class="logo-img" src="data:image/png;base64,
                {img}">
                <h1 class='logo-text'>Suzieq</h1>
            </div>
            """,
            unsafe_allow_html=True
        )

    sel_menulist = list(filter(lambda x: not x.startswith('_'), menulist))

    with srch_col:
        st.text(' ')
        search_str = st.text_input("Search", value=state.search_text,
                                   key='search', on_change=main_sync_state)
    if search_text is not None and (search_str != search_text):
        # We're assuming here that the page is titled Search
        page = 'Search'

    with page_col:
        # The empty writes are for aligning the pages link with the logo
        st.text(' ')
        set_horizontal_radio()
        st.radio('Page', sel_menulist, key='sq_page',
                 index=sel_menulist.index(page or 'Status'),
                 on_change=main_sync_state)
        page = state.sq_page

    return page, search_str


def main_sync_state():
    '''Sync search & page state on main Suzieq GUI page'''
    wsstate = st.session_state

    if wsstate.search_text != wsstate.search:
        wsstate.search_text = wsstate.search
        wsstate.page = wsstate.sq_page = 'Search'

    if wsstate.page != wsstate.sq_page:
        wsstate.page = wsstate.sq_page
        st.experimental_set_query_params(**{'page': wsstate.page})
        if wsstate.page != 'Search':
            wsstate.search_text = ''


def build_pages() -> Dict:
    '''Build dict of page name and corresponding module'''

    pages = SqGuiPage.get_plugins()
    page_tbl = defaultdict(dict)

    for obj in pages.values():
        page = obj()
        # pylint: disable=protected-access
        page_tbl[page._title] = page

    return page_tbl


# pylint: disable=too-many-statements
def apprun(*args):
    '''The main application routine'''

    if not args:
        args = sys.argv

    config_file = os.environ.get('SQ_GUI_CONFIG_FILE', '')
    config_file = sq_get_config_file(config_file)

    state = st.session_state
    state.config_file = config_file

    st.set_page_config(
        layout="wide",
        page_title="Suzieq",
        menu_items={
            'Get Help': 'https://suzieq.readthedocs.io/',
            'Report a bug':
            'https://github.com/netenglabs/suzieq/issues/new/choose',
            'About': f'''
            Suzieq Version: {SUZIEQ_VERSION}.
            Copyright 2022 Stardust Systems Inc. All Rights Reserved
            '''
        })

    for key, val in [('pages', None),
                     ('menulist', []),
                     ('page', None),
                     ('search_text', ''),
                     ('sqobjs', {})]:
        if key not in state:
            state[key] = val

    if not state.menulist:
        state.pages = build_pages()
        # Hardcoding the order of these pages
        state.menulist = [SuzieqMainPages.STATUS.value,
                          SuzieqMainPages.XPLORE.value,
                          SuzieqMainPages.PATH.value,
                          SuzieqMainPages.SEARCH.value]
        for pg, obj in state.pages.items():
            if pg not in state.menulist and obj.add_to_menu:
                state.menulist.append(pg)

    # NOTE: Do not move this before set_page_config
    if 'first_time' not in state:
        state.first_time = True

    url_params = st.experimental_get_query_params()
    if url_params.get('page', ''):
        page = url_params['page']
        if isinstance(page, list):
            page = page[0]
        old_session_state = get_main_session_by_id(
            url_params.get('session', [''])[0])
        if old_session_state:
            if page == "Path-Debug":
                state.pages[page].set_path_state(old_session_state)
                state.pages[page].build()
                st.stop()
            elif page == "Help":
                state.pages[page].build()
                st.stop()

        if page == 'Search' and state.first_time:
            search_text = url_params.get('search_text', '')
            if search_text:
                state.search_text = search_text[0]

        if isinstance(page, list):
            page = page[0]
    else:
        page = None

    if page is None:
        if state.page:
            page = state.page
        else:
            page = state.menulist[0]

    state.page = page

    page, search_text = display_title(page)
    state.first_time = False

    if search_text != state.search_text:
        state.search_text = search_text
        state.page = page = 'Search'

    state.pages[page].build()


if __name__ == '__main__':
    apprun()



import time

from nubia import command
import pandas as pd

from suzieq.cli.nubia_patch import argument
from suzieq.cli.sqcmds.command import SqCommand
from suzieq.sqobjects.ospf import OspfObj


@command("ospf", help="Act on OSPF data")
@argument(
    "ifname",
    description="Interface name(s), space separated"
)
@argument("vrf", description="VRF(s), space separated")
@argument("area", description="Area(s), space separated")
@argument("state", description="Select view based on OSPF state",
          choices=["full", "other", "passive", "!full", "!passive",
                   "!other"])
class OspfCmd(SqCommand):
    """OSPFv2 protocol information"""

    def __init__(
            self,
            engine: str = "",
            hostname: str = "",
            start_time: str = "",
            end_time: str = "",
            view: str = "",
            namespace: str = "",
            format: str = "",  # pylint: disable=redefined-builtin
            query_str: str = ' ',
            columns: str = "default",
            ifname: str = '',
            vrf: str = '',
            area: str = '',
            state: str = ''
    ) -> None:
        super().__init__(
            engine=engine,
            hostname=hostname,
            start_time=start_time,
            end_time=end_time,
            view=view,
            namespace=namespace,
            columns=columns,
            format=format,
            query_str=query_str,
            sqobj=OspfObj,
        )
        self.lvars = {
            'ifname': ifname.split(),
            'vrf': vrf.split(),
            'area': area.split(),
            'state': state
        }

    @command("assert")
    @argument("result", description="Show only assert that matches this value",
              choices=["all", "fail", "pass"])
    def aver(self, result: str = 'all') -> pd.DataFrame:
        """
        Test OSPF runtime state is without errors
        """
        now = time.time()

        df = self._invoke_sqobj(self.sqobj.aver,
                                namespace=self.namespace,
                                hostname=self.hostname,
                                result=result,
                                **self.lvars,
                                )
        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)

        return self._assert_gen_output(df)
